//=======================================================================================
// ____          ____    __    ______     __________   __      __       __        __
// \    \       |    |  |  |  |   _   \  |___    ___| |  |    |  |     /  \      |  |
//  \    \      |    |  |  |  |  |_)   |     |  |     |  |    |  |    /    \     |  |
//   \    \     |    |  |  |  |   _   /      |  |     |  |    |  |   /  /\  \    |  |
//    \    \    |    |  |  |  |  | \  \      |  |     |   \__/   |  /  ____  \   |  |____
//     \    \   |    |  |__|  |__|  \__\     |__|      \________/  /__/    \__\  |_______|
//      \    \  |    |   ________________________________________________________________
//       \    \ |    |  |  ______________________________________________________________|
//        \    \|    |  |  |         __          __     __     __     ______      _______
//         \         |  |  |_____   |  |        |  |   |  |   |  |   |   _  \    /  _____)
//          \        |  |   _____|  |  |        |  |   |  |   |  |   |  | \  \   \_______
//           \       |  |  |        |  |_____   |   \_/   |   |  |   |  |_/  /    _____  |
//            \ _____|  |__|        |________|   \_______/    |__|   |______/    (_______/
//
//  This file is part of VirtualFluids. VirtualFluids is free software: you can
//  redistribute it and/or modify it under the terms of the GNU General Public
//  License as published by the Free Software Foundation, either version 3 of
//  the License, or (at your option) any later version.
//
//  VirtualFluids is distributed in the hope that it will be useful, but WITHOUT
//  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
//  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
//  for more details.
//
//  You should have received a copy of the GNU General Public License along
//  with VirtualFluids (see COPYING.txt). If not, see <http://www.gnu.org/licenses/>.
//
//! \author Martin Schoenherr
//=======================================================================================
#ifndef GPU_INTERFACE_H
#define GPU_INTERFACE_H

#include "LBM/LB.h"

#include <cuda.h>
#include <cuda_runtime.h>

struct LBMSimulationParameter;
class Parameter;


void CalcMac27( real* vxD,
                          real* vyD,
                          real* vzD,
                          real* rhoD,
                          unsigned int* geoD,
                          unsigned int* neighborX,
                          unsigned int* neighborY,
                          unsigned int* neighborZ,
                          unsigned long long numberOfLBnodes,
                          unsigned int grid_nx, 
                          unsigned int grid_ny, 
                          unsigned int grid_nz, 
                          real* DD,
                          bool isEvenTimestep);

void CalcMacSP27(real* vxD,
                            real* vyD,
                            real* vzD,
                            real* rhoD,
                            real* pressD,
                            unsigned int* geoD,
                            unsigned int* neighborX,
                            unsigned int* neighborY,
                            unsigned int* neighborZ,
                            unsigned long long numberOfLBnodes,
                            unsigned int numberOfThreads, 
                            real* DD,
                            bool isEvenTimestep);

void CalcMacCompSP27(real* vxD,
                                real* vyD,
                                real* vzD,
                                real* rhoD,
                                real* pressD,
                                unsigned int* geoD,
                                unsigned int* neighborX,
                                unsigned int* neighborY,
                                unsigned int* neighborZ,
                                unsigned long long numberOfLBnodes,
                                unsigned int numberOfThreads, 
                                real* DD,
                                bool isEvenTimestep);

void CalcMacThS7(  real* Conc,
                              unsigned int* geoD,
                              unsigned int* neighborX,
                              unsigned int* neighborY,
                              unsigned int* neighborZ,
                              unsigned long long numberOfLBnodes,
                              unsigned int numberOfThreads, 
                              real* DD7,
                              bool isEvenTimestep);

void PlaneConcThS7(real* Conc,
                              int* kPC,
                              unsigned int numberOfPointskPC,
                              unsigned int* geoD,
                              unsigned int* neighborX,
                              unsigned int* neighborY,
                              unsigned int* neighborZ,
                              unsigned long long numberOfLBnodes,
                              unsigned int numberOfThreads, 
                              real* DD7,
                              bool isEvenTimestep);

void PlaneConcThS27(real* Conc,
                               int* kPC,
                               unsigned int numberOfPointskPC,
                               unsigned int* geoD,
                               unsigned int* neighborX,
                               unsigned int* neighborY,
                               unsigned int* neighborZ,
                               unsigned long long numberOfLBnodes,
                               unsigned int numberOfThreads, 
                               real* DD27,
                               bool isEvenTimestep);

void CalcConcentration27( unsigned int numberOfThreads,
                                     real* Conc,
                                     unsigned int* geoD,
                                     unsigned int* neighborX,
                                     unsigned int* neighborY,
                                     unsigned int* neighborZ,
                                     unsigned long long numberOfLBnodes,
                                     real* DD27,
                                     bool isEvenTimestep);

void CalcMedSP27(  real* vxD,
                              real* vyD,
                              real* vzD,
                              real* rhoD,
                              real* pressD,
                              unsigned int* geoD,
                              unsigned int* neighborX,
                              unsigned int* neighborY,
                              unsigned int* neighborZ,
                              unsigned long long numberOfLBnodes,
                              unsigned int numberOfThreads, 
                              real* DD,
                              bool isEvenTimestep);

void CalcMedCompSP27(real* vxD,
                                real* vyD,
                                real* vzD,
                                real* rhoD,
                                real* pressD,
                                unsigned int* geoD,
                                unsigned int* neighborX,
                                unsigned int* neighborY,
                                unsigned int* neighborZ,
                                unsigned long long numberOfLBnodes,
                                unsigned int numberOfThreads, 
                                real* DD,
                                bool isEvenTimestep);

void CalcMedCompAD27(
    real* vxD,
    real* vyD,
    real* vzD,
    real* rhoD,
    real* pressD,
    real* concD,
    unsigned int* geoD,
    unsigned int* neighborX,
    unsigned int* neighborY,
    unsigned int* neighborZ,
    unsigned long long numberOfLBnodes,
    unsigned int numberOfThreads,
    real* DD,
    real* DD_AD,
    bool isEvenTimestep);

void CalcMacMedSP27(  real* vxD,
                                 real* vyD,
                                 real* vzD,
                                 real* rhoD,
                                 real* pressD,
                                 unsigned int* geoD,
                                 unsigned int* neighborX,
                                 unsigned int* neighborY,
                                 unsigned int* neighborZ,
                                 unsigned int tdiff,
                                 unsigned long long numberOfLBnodes,
                                 unsigned int numberOfThreads, 
                                 bool isEvenTimestep);

void ResetMeanValuesSP27(
    real* vxD,
    real* vyD,
    real* vzD,
    real* rhoD,
    real* pressD,
    unsigned long long numberOfLBnodes,
    unsigned int numberOfThreads,
    bool isEvenTimestep);

void ResetMeanValuesAD27(
    real* vxD,
    real* vyD,
    real* vzD,
    real* rhoD,
    real* pressD,
    real* concD,
    unsigned long long numberOfLBnodes,
    unsigned int numberOfThreads,
    bool isEvenTimestep);

void Calc2ndMomentsIncompSP27(real* kxyFromfcNEQ,
                                         real* kyzFromfcNEQ,
                                         real* kxzFromfcNEQ,
                                         real* kxxMyyFromfcNEQ,
                                         real* kxxMzzFromfcNEQ,
                                         unsigned int* geoD,
                                         unsigned int* neighborX,
                                         unsigned int* neighborY,
                                         unsigned int* neighborZ,
                                         unsigned long long numberOfLBnodes,
                                         unsigned int numberOfThreads, 
                                         real* DD,
                                         bool isEvenTimestep);

void Calc2ndMomentsCompSP27(real* kxyFromfcNEQ,
                                       real* kyzFromfcNEQ,
                                       real* kxzFromfcNEQ,
                                       real* kxxMyyFromfcNEQ,
                                       real* kxxMzzFromfcNEQ,
                                       unsigned int* geoD,
                                       unsigned int* neighborX,
                                       unsigned int* neighborY,
                                       unsigned int* neighborZ,
                                       unsigned long long numberOfLBnodes,
                                       unsigned int numberOfThreads, 
                                       real* DD,
                                       bool isEvenTimestep);

void Calc3rdMomentsIncompSP27(real* CUMbbb,
                                         real* CUMabc,
                                         real* CUMbac,
                                         real* CUMbca,
                                         real* CUMcba,
                                         real* CUMacb,
                                         real* CUMcab,
                                         unsigned int* geoD,
                                         unsigned int* neighborX,
                                         unsigned int* neighborY,
                                         unsigned int* neighborZ,
                                         unsigned long long numberOfLBnodes,
                                         unsigned int numberOfThreads, 
                                         real* DD,
                                         bool isEvenTimestep);

void Calc3rdMomentsCompSP27(real* CUMbbb,
                                       real* CUMabc,
                                       real* CUMbac,
                                       real* CUMbca,
                                       real* CUMcba,
                                       real* CUMacb,
                                       real* CUMcab,
                                       unsigned int* geoD,
                                       unsigned int* neighborX,
                                       unsigned int* neighborY,
                                       unsigned int* neighborZ,
                                       unsigned long long numberOfLBnodes,
                                       unsigned int numberOfThreads, 
                                       real* DD,
                                       bool isEvenTimestep);

void CalcHigherMomentsIncompSP27(real* CUMcbb,
                                            real* CUMbcb,
                                            real* CUMbbc,
                                            real* CUMcca,
                                            real* CUMcac,
                                            real* CUMacc,
                                            real* CUMbcc,
                                            real* CUMcbc,
                                            real* CUMccb,
                                            real* CUMccc,
                                            unsigned int* geoD,
                                            unsigned int* neighborX,
                                            unsigned int* neighborY,
                                            unsigned int* neighborZ,
                                            unsigned long long numberOfLBnodes,
                                            unsigned int numberOfThreads, 
                                            real* DD,
                                            bool isEvenTimestep);

void CalcHigherMomentsCompSP27(real* CUMcbb,
                                          real* CUMbcb,
                                          real* CUMbbc,
                                          real* CUMcca,
                                          real* CUMcac,
                                          real* CUMacc,
                                          real* CUMbcc,
                                          real* CUMcbc,
                                          real* CUMccb,
                                          real* CUMccc,
                                          unsigned int* geoD,
                                          unsigned int* neighborX,
                                          unsigned int* neighborY,
                                          unsigned int* neighborZ,
                                          unsigned long long numberOfLBnodes,
                                          unsigned int numberOfThreads, 
                                          real* DD,
                                          bool isEvenTimestep);

void LBCalcMeasurePoints27(real* vxMP,
                                      real* vyMP,
                                      real* vzMP,
                                      real* rhoMP,
                                      unsigned int* kMP,
                                      unsigned int numberOfPointskMP,
                                      unsigned int MPClockCycle,
                                      unsigned int t,
                                      unsigned int* geoD,
                                      unsigned int* neighborX,
                                      unsigned int* neighborY,
                                      unsigned int* neighborZ,
                                      unsigned long long numberOfLBnodes,
                                      real* DD,
                                      unsigned int numberOfThreads, 
                                      bool isEvenTimestep);

void QADDev7(unsigned int numberOfThreads,
                        real* DD, 
                        real* DD7,
                        real* temp,
                        real diffusivity,
                        int* k_Q, 
                        real* QQ,
                        unsigned int numberOfBCnodes, 
                        real om1, 
                        unsigned int* neighborX,
                        unsigned int* neighborY,
                        unsigned int* neighborZ,
                        unsigned long long numberOfLBnodes, 
                        bool isEvenTimestep);

//////////////////////////////////////////////////////////////////////////
//! \brief defines the behavior of a slip-AD boundary condition
void ADSlipVelDevComp(
    uint numberOfThreads,
    real * normalX,
    real * normalY,
    real * normalZ,
    real * distributions,
    real * distributionsAD,
    int* QindexArray,
    real * Qarrays,
    uint numberOfBCnodes,
    real omegaDiffusivity,
    uint * neighborX,
    uint * neighborY,
    uint * neighborZ,
    unsigned long long numberOfLBnodes,
    bool isEvenTimestep);
    
void QADDirichletDev27( unsigned int numberOfThreads,
                                   real* DD, 
                                   real* DD27,
                                   real* temp,
                                   real diffusivity,
                                   int* k_Q, 
                                   real* QQ,
                                   unsigned int numberOfBCnodes, 
                                   real om1, 
                                   unsigned int* neighborX,
                                   unsigned int* neighborY,
                                   unsigned int* neighborZ,
                                   unsigned long long numberOfLBnodes, 
                                   bool isEvenTimestep);

void QADBBDev27(  unsigned int numberOfThreads,
                             real* DD, 
                             real* DD27,
                             real* temp,
                             real diffusivity,
                             int* k_Q, 
                             real* QQ,
                             unsigned int numberOfBCnodes, 
                             real om1, 
                             unsigned int* neighborX,
                             unsigned int* neighborY,
                             unsigned int* neighborZ,
                             unsigned long long numberOfLBnodes, 
                             bool isEvenTimestep);

void QADVelDev7(unsigned int numberOfThreads,
                           real* DD, 
                           real* DD7,
                           real* temp,
                           real* velo,
                           real diffusivity,
                           int* k_Q, 
                           real* QQ,
                           unsigned int numberOfBCnodes, 
                           real om1, 
                           unsigned int* neighborX,
                           unsigned int* neighborY,
                           unsigned int* neighborZ,
                           unsigned long long numberOfLBnodes, 
                           bool isEvenTimestep);


void QADVelDev27(  unsigned int numberOfThreads,
                              real* DD, 
                              real* DD27,
                              real* temp,
                              real* velo,
                              real diffusivity,
                              int* k_Q, 
                              real* QQ,
                              unsigned int numberOfBCnodes, 
                              real om1, 
                              unsigned int* neighborX,
                              unsigned int* neighborY,
                              unsigned int* neighborZ,
                              unsigned long long numberOfLBnodes, 
                              bool isEvenTimestep);

void QADPressDev7( unsigned int numberOfThreads,
                              real* DD, 
                              real* DD7,
                              real* temp,
                              real* velo,
                              real diffusivity,
                              int* k_Q, 
                              real* QQ,
                              unsigned int numberOfBCnodes, 
                              real om1, 
                              unsigned int* neighborX,
                              unsigned int* neighborY,
                              unsigned int* neighborZ,
                              unsigned long long numberOfLBnodes, 
                              bool isEvenTimestep);

void QADPressDev27(unsigned int numberOfThreads,
                              real* DD, 
                              real* DD27,
                              real* temp,
                              real* velo,
                              real diffusivity,
                              int* k_Q, 
                              real* QQ,
                              unsigned int numberOfBCnodes, 
                              real om1, 
                              unsigned int* neighborX,
                              unsigned int* neighborY,
                              unsigned int* neighborZ,
                              unsigned long long numberOfLBnodes, 
                              bool isEvenTimestep);

void QADPressNEQNeighborDev27(
                                            unsigned int numberOfThreads,
                                            real* DD,
                                            real* DD27,
                                            int* k_Q,
                                            int* k_N,
                                            int numberOfBCnodes,
                                            unsigned int* neighborX,
                                            unsigned int* neighborY,
                                            unsigned int* neighborZ,
                                            unsigned long long numberOfLBnodes,
                                            bool isEvenTimestep
                                        );

void QNoSlipADincompDev7(unsigned int numberOfThreads,
                                    real* DD, 
                                    real* DD7,
                                    real* temp,
                                    real diffusivity,
                                    int* k_Q, 
                                    real* QQ,
                                    unsigned int numberOfBCnodes, 
                                    real om1, 
                                    unsigned int* neighborX,
                                    unsigned int* neighborY,
                                    unsigned int* neighborZ,
                                    unsigned long long numberOfLBnodes, 
                                    bool isEvenTimestep);

void QNoSlipADincompDev27(unsigned int numberOfThreads,
                                     real* DD, 
                                     real* DD27,
                                     real* temp,
                                     real diffusivity,
                                     int* k_Q, 
                                     real* QQ,
                                     unsigned int numberOfBCnodes, 
                                     real om1, 
                                     unsigned int* neighborX,
                                     unsigned int* neighborY,
                                     unsigned int* neighborZ,
                                     unsigned long long numberOfLBnodes, 
                                     bool isEvenTimestep);

void QADVeloIncompDev7( unsigned int numberOfThreads,
                                   real* DD, 
                                   real* DD7,
                                   real* temp,
                                   real* velo,
                                   real diffusivity,
                                   int* k_Q, 
                                   real* QQ,
                                   unsigned int numberOfBCnodes, 
                                   real om1, 
                                   unsigned int* neighborX,
                                   unsigned int* neighborY,
                                   unsigned int* neighborZ,
                                   unsigned long long numberOfLBnodes, 
                                   bool isEvenTimestep);


void QADVeloIncompDev27( unsigned int numberOfThreads,
                                    real* DD, 
                                    real* DD27,
                                    real* temp,
                                    real* velo,
                                    real diffusivity,
                                    int* k_Q, 
                                    real* QQ,
                                    unsigned int numberOfBCnodes, 
                                    real om1, 
                                    unsigned int* neighborX,
                                    unsigned int* neighborY,
                                    unsigned int* neighborZ,
                                    unsigned long long numberOfLBnodes, 
                                    bool isEvenTimestep);

void QADPressIncompDev7(  unsigned int numberOfThreads,
                                     real* DD, 
                                     real* DD7,
                                     real* temp,
                                     real* velo,
                                     real diffusivity,
                                     int* k_Q, 
                                     real* QQ,
                                     unsigned int numberOfBCnodes, 
                                     real om1, 
                                     unsigned int* neighborX,
                                     unsigned int* neighborY,
                                     unsigned int* neighborZ,
                                     unsigned long long numberOfLBnodes, 
                                     bool isEvenTimestep);

void QADPressIncompDev27(  unsigned int numberOfThreads,
                                      real* DD, 
                                      real* DD27,
                                      real* temp,
                                      real* velo,
                                      real diffusivity,
                                      int* k_Q, 
                                      real* QQ,
                                      unsigned int numberOfBCnodes, 
                                      real om1, 
                                      unsigned int* neighborX,
                                      unsigned int* neighborY,
                                      unsigned int* neighborZ,
                                      unsigned long long numberOfLBnodes, 
                                      bool isEvenTimestep);

void ScaleCF27( real* DC, 
                           real* DF, 
                           unsigned int* neighborCX,
                           unsigned int* neighborCY,
                           unsigned int* neighborCZ,
                           unsigned int* neighborFX,
                           unsigned int* neighborFY,
                           unsigned int* neighborFZ,
                           unsigned long long numberOfLBnodesC, 
                           unsigned long long numberOfLBnodesF, 
                           bool isEvenTimestep,
                           unsigned int* posCSWB, 
                           unsigned int* posFSWB, 
                           unsigned int kCF, 
                           real omCoarse, 
                           real omFine, 
                           real nu, 
                           unsigned int nxC, 
                           unsigned int nyC, 
                           unsigned int nxF, 
                           unsigned int nyF,
                           unsigned int numberOfThreads);

void ScaleFC27( real* DC, 
                           real* DF, 
                           unsigned int* neighborCX,
                           unsigned int* neighborCY,
                           unsigned int* neighborCZ,
                           unsigned int* neighborFX,
                           unsigned int* neighborFY,
                           unsigned int* neighborFZ,
                           unsigned long long numberOfLBnodesC, 
                           unsigned long long numberOfLBnodesF, 
                           bool isEvenTimestep,
                           unsigned int* posC, 
                           unsigned int* posFSWB, 
                           unsigned int kFC, 
                           real omCoarse, 
                           real omFine, 
                           real nu, 
                           unsigned int nxC, 
                           unsigned int nyC, 
                           unsigned int nxF, 
                           unsigned int nyF,
                           unsigned int numberOfThreads);

void ScaleCFEff27(real* DC, 
                             real* DF, 
                             unsigned int* neighborCX,
                             unsigned int* neighborCY,
                             unsigned int* neighborCZ,
                             unsigned int* neighborFX,
                             unsigned int* neighborFY,
                             unsigned int* neighborFZ,
                             unsigned long long numberOfLBnodesC, 
                             unsigned long long numberOfLBnodesF, 
                             bool isEvenTimestep,
                             unsigned int* posCSWB, 
                             unsigned int* posFSWB, 
                             unsigned int kCF, 
                             real omCoarse, 
                             real omFine, 
                             real nu, 
                             unsigned int nxC, 
                             unsigned int nyC, 
                             unsigned int nxF, 
                             unsigned int nyF,
                             unsigned int numberOfThreads,
                             ICellNeigh neighborCoarseToFine);

void ScaleFCEff27(real* DC, 
                             real* DF, 
                             unsigned int* neighborCX,
                             unsigned int* neighborCY,
                             unsigned int* neighborCZ,
                             unsigned int* neighborFX,
                             unsigned int* neighborFY,
                             unsigned int* neighborFZ,
                             unsigned long long numberOfLBnodesC, 
                             unsigned long long numberOfLBnodesF, 
                             bool isEvenTimestep,
                             unsigned int* posC, 
                             unsigned int* posFSWB, 
                             unsigned int kFC, 
                             real omCoarse, 
                             real omFine, 
                             real nu, 
                             unsigned int nxC, 
                             unsigned int nyC, 
                             unsigned int nxF, 
                             unsigned int nyF,
                             unsigned int numberOfThreads,
                             ICellNeigh neighborFineToCoarse);

void ScaleCFLast27(real* DC, 
                              real* DF, 
                              unsigned int* neighborCX,
                              unsigned int* neighborCY,
                              unsigned int* neighborCZ,
                              unsigned int* neighborFX,
                              unsigned int* neighborFY,
                              unsigned int* neighborFZ,
                              unsigned long long numberOfLBnodesC, 
                              unsigned long long numberOfLBnodesF, 
                              bool isEvenTimestep,
                              unsigned int* posCSWB, 
                              unsigned int* posFSWB, 
                              unsigned int kCF, 
                              real omCoarse, 
                              real omFine, 
                              real nu, 
                              unsigned int nxC, 
                              unsigned int nyC, 
                              unsigned int nxF, 
                              unsigned int nyF,
                              unsigned int numberOfThreads,
                              ICellNeigh neighborCoarseToFine);

void ScaleFCLast27(real* DC, 
                              real* DF, 
                              unsigned int* neighborCX,
                              unsigned int* neighborCY,
                              unsigned int* neighborCZ,
                              unsigned int* neighborFX,
                              unsigned int* neighborFY,
                              unsigned int* neighborFZ,
                              unsigned long long numberOfLBnodesC, 
                              unsigned long long numberOfLBnodesF, 
                              bool isEvenTimestep,
                              unsigned int* posC, 
                              unsigned int* posFSWB, 
                              unsigned int kFC, 
                              real omCoarse, 
                              real omFine, 
                              real nu, 
                              unsigned int nxC, 
                              unsigned int nyC, 
                              unsigned int nxF, 
                              unsigned int nyF,
                              unsigned int numberOfThreads,
                              ICellNeigh neighborFineToCoarse);

void ScaleCFpress27(real* DC, 
                              real* DF, 
                              unsigned int* neighborCX,
                              unsigned int* neighborCY,
                              unsigned int* neighborCZ,
                              unsigned int* neighborFX,
                              unsigned int* neighborFY,
                              unsigned int* neighborFZ,
                              unsigned long long numberOfLBnodesC, 
                              unsigned long long numberOfLBnodesF, 
                              bool isEvenTimestep,
                              unsigned int* posCSWB, 
                              unsigned int* posFSWB, 
                              unsigned int kCF, 
                              real omCoarse, 
                              real omFine, 
                              real nu, 
                              unsigned int nxC, 
                              unsigned int nyC, 
                              unsigned int nxF, 
                              unsigned int nyF,
                              unsigned int numberOfThreads,
                              ICellNeigh neighborCoarseToFine);

void ScaleFCpress27(  real* DC, 
                                 real* DF, 
                                 unsigned int* neighborCX,
                                 unsigned int* neighborCY,
                                 unsigned int* neighborCZ,
                                 unsigned int* neighborFX,
                                 unsigned int* neighborFY,
                                 unsigned int* neighborFZ,
                                 unsigned long long numberOfLBnodesC, 
                                 unsigned long long numberOfLBnodesF, 
                                 bool isEvenTimestep,
                                 unsigned int* posC, 
                                 unsigned int* posFSWB, 
                                 unsigned int kFC, 
                                 real omCoarse, 
                                 real omFine, 
                                 real nu, 
                                 unsigned int nxC, 
                                 unsigned int nyC, 
                                 unsigned int nxF, 
                                 unsigned int nyF,
                                 unsigned int numberOfThreads,
                                 ICellNeigh neighborFineToCoarse);

void ScaleCF_Fix_27(real* DC, 
                              real* DF, 
                              unsigned int* neighborCX,
                              unsigned int* neighborCY,
                              unsigned int* neighborCZ,
                              unsigned int* neighborFX,
                              unsigned int* neighborFY,
                              unsigned int* neighborFZ,
                              unsigned long long numberOfLBnodesC, 
                              unsigned long long numberOfLBnodesF, 
                              bool isEvenTimestep,
                              unsigned int* posCSWB, 
                              unsigned int* posFSWB, 
                              unsigned int kCF, 
                              real omCoarse, 
                              real omFine, 
                              real nu, 
                              unsigned int nxC, 
                              unsigned int nyC, 
                              unsigned int nxF, 
                              unsigned int nyF,
                              unsigned int numberOfThreads,
                              ICellNeigh neighborCoarseToFine);

void ScaleCF_Fix_comp_27(   real* DC, 
                                       real* DF, 
                                       unsigned int* neighborCX,
                                       unsigned int* neighborCY,
                                       unsigned int* neighborCZ,
                                       unsigned int* neighborFX,
                                       unsigned int* neighborFY,
                                       unsigned int* neighborFZ,
                                       unsigned long long numberOfLBnodesC, 
                                       unsigned long long numberOfLBnodesF, 
                                       bool isEvenTimestep,
                                       unsigned int* posCSWB, 
                                       unsigned int* posFSWB, 
                                       unsigned int kCF, 
                                       real omCoarse, 
                                       real omFine, 
                                       real nu, 
                                       unsigned int nxC, 
                                       unsigned int nyC, 
                                       unsigned int nxF, 
                                       unsigned int nyF,
                                       unsigned int numberOfThreads,
                                       ICellNeigh neighborCoarseToFine);

void ScaleCF_0817_comp_27(  real* DC, 
                                       real* DF, 
                                       unsigned int* neighborCX,
                                       unsigned int* neighborCY,
                                       unsigned int* neighborCZ,
                                       unsigned int* neighborFX,
                                       unsigned int* neighborFY,
                                       unsigned int* neighborFZ,
                                       unsigned long long numberOfLBnodesC, 
                                       unsigned long long numberOfLBnodesF, 
                                       bool isEvenTimestep,
                                       unsigned int* posCSWB, 
                                       unsigned int* posFSWB, 
                                       unsigned int kCF, 
                                       real omCoarse, 
                                       real omFine, 
                                       real nu, 
                                       unsigned int nxC, 
                                       unsigned int nyC, 
                                       unsigned int nxF, 
                                       unsigned int nyF,
                                       unsigned int numberOfThreads,
                                       ICellNeigh neighborCoarseToFine,
                                       CUstream_st* stream);

void ScaleCF_comp_D3Q27F3_2018(    real* DC,
                                            real* DF,
                                            real* G6, 
                                            unsigned int* neighborCX,
                                            unsigned int* neighborCY,
                                            unsigned int* neighborCZ,
                                            unsigned int* neighborFX,
                                            unsigned int* neighborFY,
                                            unsigned int* neighborFZ,
                                            unsigned long long numberOfLBnodesC, 
                                            unsigned long long numberOfLBnodesF, 
                                            bool isEvenTimestep,
                                            unsigned int* posCSWB, 
                                            unsigned int* posFSWB, 
                                            unsigned int kCF, 
                                            real omCoarse, 
                                            real omFine, 
                                            real nu, 
                                            unsigned int nxC, 
                                            unsigned int nyC, 
                                            unsigned int nxF, 
                                            unsigned int nyF,
                                            unsigned int numberOfThreads,
                                            ICellNeigh neighborCoarseToFine);

void ScaleCF_comp_D3Q27F3(real* DC,
                                     real* DF,
                                     real* G6, 
                                     unsigned int* neighborCX,
                                     unsigned int* neighborCY,
                                     unsigned int* neighborCZ,
                                     unsigned int* neighborFX,
                                     unsigned int* neighborFY,
                                     unsigned int* neighborFZ,
                                     unsigned long long numberOfLBnodesC, 
                                     unsigned long long numberOfLBnodesF, 
                                     bool isEvenTimestep,
                                     unsigned int* posCSWB, 
                                     unsigned int* posFSWB, 
                                     unsigned int kCF, 
                                     real omCoarse, 
                                     real omFine, 
                                     real nu, 
                                     unsigned int nxC, 
                                     unsigned int nyC, 
                                     unsigned int nxF, 
                                     unsigned int nyF,
                                     unsigned int numberOfThreads,
                                     ICellNeigh neighborCoarseToFine,
                                     CUstream_st *stream);

void ScaleCF_staggered_time_comp_27( real* DC, 
                                                real* DF, 
                                                unsigned int* neighborCX,
                                                unsigned int* neighborCY,
                                                unsigned int* neighborCZ,
                                                unsigned int* neighborFX,
                                                unsigned int* neighborFY,
                                                unsigned int* neighborFZ,
                                                unsigned long long numberOfLBnodesC, 
                                                unsigned long long numberOfLBnodesF, 
                                                bool isEvenTimestep,
                                                unsigned int* posCSWB, 
                                                unsigned int* posFSWB, 
                                                unsigned int kCF, 
                                                real omCoarse, 
                                                real omFine, 
                                                real nu, 
                                                unsigned int nxC, 
                                                unsigned int nyC, 
                                                unsigned int nxF, 
                                                unsigned int nyF,
                                                unsigned int numberOfThreads,
                                                ICellNeigh neighborCoarseToFine);

void ScaleCF_RhoSq_comp_27(LBMSimulationParameter * parameterDeviceC, LBMSimulationParameter* parameterDeviceF, ICells * interpolationCellsCoarseToFine, ICellNeigh &neighborCoarseToFine, CUstream_st *stream);

template<bool hasTurbulentViscosity> void ScaleCF_compressible(LBMSimulationParameter * parameterDeviceC, LBMSimulationParameter* parameterDeviceF, ICells * interpolationCellsCoarseToFine, ICellNeigh &neighborCoarseToFine, CUstream_st *stream);

void ScaleCF_RhoSq_3rdMom_comp_27( real* DC, 
                                              real* DF, 
                                              unsigned int* neighborCX,
                                              unsigned int* neighborCY,
                                              unsigned int* neighborCZ,
                                              unsigned int* neighborFX,
                                              unsigned int* neighborFY,
                                              unsigned int* neighborFZ,
                                              unsigned long long numberOfLBnodesC, 
                                              unsigned long long numberOfLBnodesF, 
                                              bool isEvenTimestep,
                                              unsigned int* posCSWB, 
                                              unsigned int* posFSWB, 
                                              unsigned int kCF, 
                                              real omCoarse, 
                                              real omFine, 
                                              real nu, 
                                              unsigned int nxC, 
                                              unsigned int nyC, 
                                              unsigned int nxF, 
                                              unsigned int nyF,
                                              unsigned int numberOfThreads,
                                              ICellNeigh neighborCoarseToFine,
                                              CUstream_st *stream);

void ScaleCF_AA2016_comp_27( real* DC, 
                                        real* DF, 
                                        unsigned int* neighborCX,
                                        unsigned int* neighborCY,
                                        unsigned int* neighborCZ,
                                        unsigned int* neighborFX,
                                        unsigned int* neighborFY,
                                        unsigned int* neighborFZ,
                                        unsigned long long numberOfLBnodesC, 
                                        unsigned long long numberOfLBnodesF, 
                                        bool isEvenTimestep,
                                        unsigned int* posCSWB, 
                                        unsigned int* posFSWB, 
                                        unsigned int kCF, 
                                        real omCoarse, 
                                        real omFine, 
                                        real nu, 
                                        unsigned int nxC, 
                                        unsigned int nyC, 
                                        unsigned int nxF, 
                                        unsigned int nyF,
                                        unsigned int numberOfThreads,
                                        ICellNeigh neighborCoarseToFine,
                                        CUstream_st *stream);

void ScaleCF_NSPress_27(real* DC, 
                                  real* DF, 
                                  unsigned int* neighborCX,
                                  unsigned int* neighborCY,
                                  unsigned int* neighborCZ,
                                  unsigned int* neighborFX,
                                  unsigned int* neighborFY,
                                  unsigned int* neighborFZ,
                                  unsigned long long numberOfLBnodesC, 
                                  unsigned long long numberOfLBnodesF, 
                                  bool isEvenTimestep,
                                  unsigned int* posCSWB, 
                                  unsigned int* posFSWB, 
                                  unsigned int kCF, 
                                  real omCoarse, 
                                  real omFine, 
                                  real nu, 
                                  unsigned int nxC, 
                                  unsigned int nyC, 
                                  unsigned int nxF, 
                                  unsigned int nyF,
                                  unsigned int numberOfThreads,
                                  ICellNeigh neighborCoarseToFine);

void ScaleFC_Fix_27(  real* DC, 
                                 real* DF, 
                                 unsigned int* neighborCX,
                                 unsigned int* neighborCY,
                                 unsigned int* neighborCZ,
                                 unsigned int* neighborFX,
                                 unsigned int* neighborFY,
                                 unsigned int* neighborFZ,
                                 unsigned long long numberOfLBnodesC, 
                                 unsigned long long numberOfLBnodesF, 
                                 bool isEvenTimestep,
                                 unsigned int* posC, 
                                 unsigned int* posFSWB, 
                                 unsigned int kFC, 
                                 real omCoarse, 
                                 real omFine, 
                                 real nu, 
                                 unsigned int nxC, 
                                 unsigned int nyC, 
                                 unsigned int nxF, 
                                 unsigned int nyF,
                                 unsigned int numberOfThreads,
                                 ICellNeigh neighborFineToCoarse);

void ScaleFC_Fix_comp_27(   real* DC, 
                                       real* DF, 
                                       unsigned int* neighborCX,
                                       unsigned int* neighborCY,
                                       unsigned int* neighborCZ,
                                       unsigned int* neighborFX,
                                       unsigned int* neighborFY,
                                       unsigned int* neighborFZ,
                                       unsigned long long numberOfLBnodesC, 
                                       unsigned long long numberOfLBnodesF, 
                                       bool isEvenTimestep,
                                       unsigned int* posC, 
                                       unsigned int* posFSWB, 
                                       unsigned int kFC, 
                                       real omCoarse, 
                                       real omFine, 
                                       real nu, 
                                       unsigned int nxC, 
                                       unsigned int nyC, 
                                       unsigned int nxF, 
                                       unsigned int nyF,
                                       unsigned int numberOfThreads,
                                       ICellNeigh neighborFineToCoarse);

void ScaleFC_0817_comp_27(  real* DC, 
                                       real* DF, 
                                       unsigned int* neighborCX,
                                       unsigned int* neighborCY,
                                       unsigned int* neighborCZ,
                                       unsigned int* neighborFX,
                                       unsigned int* neighborFY,
                                       unsigned int* neighborFZ,
                                       unsigned long long numberOfLBnodesC, 
                                       unsigned long long numberOfLBnodesF, 
                                       bool isEvenTimestep,
                                       unsigned int* posC, 
                                       unsigned int* posFSWB, 
                                       unsigned int kFC, 
                                       real omCoarse, 
                                       real omFine, 
                                       real nu, 
                                       unsigned int nxC, 
                                       unsigned int nyC, 
                                       unsigned int nxF, 
                                       unsigned int nyF,
                                       unsigned int numberOfThreads,
                                       ICellNeigh neighborFineToCoarse,
                                       CUstream_st *stream);

void ScaleFC_comp_D3Q27F3_2018(real* DC,
                                          real* DF,
                                          real* G6,
                                          unsigned int* neighborCX,
                                          unsigned int* neighborCY,
                                          unsigned int* neighborCZ,
                                          unsigned int* neighborFX,
                                          unsigned int* neighborFY,
                                          unsigned int* neighborFZ,
                                          unsigned long long numberOfLBnodesC, 
                                          unsigned long long numberOfLBnodesF, 
                                          bool isEvenTimestep,
                                          unsigned int* posC, 
                                          unsigned int* posFSWB, 
                                          unsigned int kFC, 
                                          real omCoarse, 
                                          real omFine, 
                                          real nu, 
                                          unsigned int nxC, 
                                          unsigned int nyC, 
                                          unsigned int nxF, 
                                          unsigned int nyF,
                                          unsigned int numberOfThreads,
                                          ICellNeigh neighborFineToCoarse);

void ScaleFC_comp_D3Q27F3( real* DC,
                                      real* DF,
                                      real* G6,
                                      unsigned int* neighborCX,
                                      unsigned int* neighborCY,
                                      unsigned int* neighborCZ,
                                      unsigned int* neighborFX,
                                      unsigned int* neighborFY,
                                      unsigned int* neighborFZ,
                                      unsigned long long numberOfLBnodesC, 
                                      unsigned long long numberOfLBnodesF, 
                                      bool isEvenTimestep,
                                      unsigned int* posC, 
                                      unsigned int* posFSWB, 
                                      unsigned int kFC, 
                                      real omCoarse, 
                                      real omFine, 
                                      real nu, 
                                      unsigned int nxC, 
                                      unsigned int nyC, 
                                      unsigned int nxF, 
                                      unsigned int nyF,
                                      unsigned int numberOfThreads,
                                      ICellNeigh neighborFineToCoarse,
                                      CUstream_st *stream);

void ScaleFC_staggered_time_comp_27( real* DC, 
                                                real* DF, 
                                                unsigned int* neighborCX,
                                                unsigned int* neighborCY,
                                                unsigned int* neighborCZ,
                                                unsigned int* neighborFX,
                                                unsigned int* neighborFY,
                                                unsigned int* neighborFZ,
                                                unsigned long long numberOfLBnodesC, 
                                                unsigned long long numberOfLBnodesF, 
                                                bool isEvenTimestep,
                                                unsigned int* posC, 
                                                unsigned int* posFSWB, 
                                                unsigned int kFC, 
                                                real omCoarse, 
                                                real omFine, 
                                                real nu, 
                                                unsigned int nxC, 
                                                unsigned int nyC, 
                                                unsigned int nxF, 
                                                unsigned int nyF,
                                                unsigned int numberOfThreads,
                                                ICellNeigh neighborFineToCoarse);

void ScaleFC_RhoSq_comp_27(LBMSimulationParameter * parameterDeviceC, LBMSimulationParameter* parameterDeviceF, ICells * interpolationCellsFineToCoarse, ICellNeigh& neighborFineToCoarse, CUstream_st *stream);

template<bool hasTurbulentViscosity> void ScaleFC_compressible(LBMSimulationParameter * parameterDeviceC, LBMSimulationParameter* parameterDeviceF, ICells * icellFC, ICellNeigh& neighborFineToCoarse, CUstream_st *stream);

void ScaleFC_RhoSq_3rdMom_comp_27( real* DC, 
                                              real* DF, 
                                              unsigned int* neighborCX,
                                              unsigned int* neighborCY,
                                              unsigned int* neighborCZ,
                                              unsigned int* neighborFX,
                                              unsigned int* neighborFY,
                                              unsigned int* neighborFZ,
                                              unsigned long long numberOfLBnodesC, 
                                              unsigned long long numberOfLBnodesF, 
                                              bool isEvenTimestep,
                                              unsigned int* posC, 
                                              unsigned int* posFSWB, 
                                              unsigned int kFC, 
                                              real omCoarse, 
                                              real omFine, 
                                              real nu, 
                                              unsigned int nxC, 
                                              unsigned int nyC, 
                                              unsigned int nxF, 
                                              unsigned int nyF,
                                              unsigned int numberOfThreads,
                                              ICellNeigh neighborFineToCoarse,
                                              CUstream_st *stream);

void ScaleFC_AA2016_comp_27( real* DC, 
                                        real* DF, 
                                        unsigned int* neighborCX,
                                        unsigned int* neighborCY,
                                        unsigned int* neighborCZ,
                                        unsigned int* neighborFX,
                                        unsigned int* neighborFY,
                                        unsigned int* neighborFZ,
                                        unsigned long long numberOfLBnodesC, 
                                        unsigned long long numberOfLBnodesF, 
                                        bool isEvenTimestep,
                                        unsigned int* posC, 
                                        unsigned int* posFSWB, 
                                        unsigned int kFC, 
                                        real omCoarse, 
                                        real omFine, 
                                        real nu, 
                                        unsigned int nxC, 
                                        unsigned int nyC, 
                                        unsigned int nxF, 
                                        unsigned int nyF,
                                        unsigned int numberOfThreads,
                                        ICellNeigh neighborFineToCoarse,
                                        CUstream_st *stream);

void ScaleFC_NSPress_27(  real* DC, 
                                     real* DF, 
                                     unsigned int* neighborCX,
                                     unsigned int* neighborCY,
                                     unsigned int* neighborCZ,
                                     unsigned int* neighborFX,
                                     unsigned int* neighborFY,
                                     unsigned int* neighborFZ,
                                     unsigned long long numberOfLBnodesC, 
                                     unsigned long long numberOfLBnodesF, 
                                     bool isEvenTimestep,
                                     unsigned int* posC, 
                                     unsigned int* posFSWB, 
                                     unsigned int kFC, 
                                     real omCoarse, 
                                     real omFine, 
                                     real nu, 
                                     unsigned int nxC, 
                                     unsigned int nyC, 
                                     unsigned int nxF, 
                                     unsigned int nyF,
                                     unsigned int numberOfThreads,
                                     ICellNeigh neighborFineToCoarse);

void ScaleCFThS7(  real* DC, 
                              real* DF, 
                              real* DD7C, 
                              real* DD7F,
                              unsigned int* neighborCX,
                              unsigned int* neighborCY,
                              unsigned int* neighborCZ,
                              unsigned int* neighborFX,
                              unsigned int* neighborFY,
                              unsigned int* neighborFZ,
                              unsigned long long numberOfLBnodesC, 
                              unsigned long long numberOfLBnodesF, 
                              bool isEvenTimestep,
                              unsigned int* posCSWB, 
                              unsigned int* posFSWB, 
                              unsigned int kCF, 
                              real nu,
                              real diffusivity_fine,
                              unsigned int numberOfThreads);

void ScaleFCThS7(  real* DC, 
                              real* DF,
                              real* DD7C, 
                              real* DD7F,
                              unsigned int* neighborCX,
                              unsigned int* neighborCY,
                              unsigned int* neighborCZ,
                              unsigned int* neighborFX,
                              unsigned int* neighborFY,
                              unsigned int* neighborFZ,
                              unsigned long long numberOfLBnodesC, 
                              unsigned long long numberOfLBnodesF, 
                              bool isEvenTimestep,
                              unsigned int* posC, 
                              unsigned int* posFSWB, 
                              unsigned int kFC, 
                              real nu,
                              real diffusivity_coarse,
                              unsigned int numberOfThreads);

void ScaleCFThSMG7(   real* DC, 
                                 real* DF,
                                 real* DD7C, 
                                 real* DD7F,
                                 unsigned int* neighborCX,
                                 unsigned int* neighborCY,
                                 unsigned int* neighborCZ,
                                 unsigned int* neighborFX,
                                 unsigned int* neighborFY,
                                 unsigned int* neighborFZ,
                                 unsigned long long numberOfLBnodesC, 
                                 unsigned long long numberOfLBnodesF, 
                                 bool isEvenTimestep,
                                 unsigned int* posCSWB, 
                                 unsigned int* posFSWB, 
                                 unsigned int kCF, 
                                 real nu,
                                 real diffusivity_fine,
                                 unsigned int numberOfThreads,
                                 ICellNeigh neighborCoarseToFine);

void ScaleFCThSMG7(real* DC, 
                              real* DF,
                              real* DD7C, 
                              real* DD7F,
                              unsigned int* neighborCX,
                              unsigned int* neighborCY,
                              unsigned int* neighborCZ,
                              unsigned int* neighborFX,
                              unsigned int* neighborFY,
                              unsigned int* neighborFZ,
                              unsigned long long numberOfLBnodesC, 
                              unsigned long long numberOfLBnodesF, 
                              bool isEvenTimestep,
                              unsigned int* posC, 
                              unsigned int* posFSWB, 
                              unsigned int kFC, 
                              real nu,
                              real diffusivity_coarse,
                              unsigned int numberOfThreads,
                              ICellNeigh neighborFineToCoarse);

void ScaleCFThS27( real* DC, 
                              real* DF, 
                              real* DD27C, 
                              real* DD27F,
                              unsigned int* neighborCX,
                              unsigned int* neighborCY,
                              unsigned int* neighborCZ,
                              unsigned int* neighborFX,
                              unsigned int* neighborFY,
                              unsigned int* neighborFZ,
                              unsigned long long numberOfLBnodesC, 
                              unsigned long long numberOfLBnodesF, 
                              bool isEvenTimestep,
                              unsigned int* posCSWB, 
                              unsigned int* posFSWB, 
                              unsigned int kCF, 
                              real nu,
                              real diffusivity_fine,
                              unsigned int numberOfThreads,
                              ICellNeigh neighborCoarseToFine);

void ScaleFCThS27( real* DC, 
                              real* DF,
                              real* DD27C, 
                              real* DD27F,
                              unsigned int* neighborCX,
                              unsigned int* neighborCY,
                              unsigned int* neighborCZ,
                              unsigned int* neighborFX,
                              unsigned int* neighborFY,
                              unsigned int* neighborFZ,
                              unsigned long long numberOfLBnodesC, 
                              unsigned long long numberOfLBnodesF, 
                              bool isEvenTimestep,
                              unsigned int* posC, 
                              unsigned int* posFSWB, 
                              unsigned int kFC, 
                              real nu,
                              real diffusivity_coarse,
                              unsigned int numberOfThreads,
                              ICellNeigh neighborFineToCoarse);

void DragLiftPostD27(real* DD, 
                                int* k_Q, 
                                real* QQ,
                                int numberOfBCnodes, 
                                double *DragX,
                                double *DragY,
                                double *DragZ,
                                unsigned int* neighborX,
                                unsigned int* neighborY,
                                unsigned int* neighborZ,
                                unsigned long long numberOfLBnodes, 
                                bool isEvenTimestep,
                                unsigned int numberOfThreads);

void DragLiftPreD27( real* DD, 
                                int* k_Q, 
                                real* QQ,
                                int numberOfBCnodes, 
                                double *DragX,
                                double *DragY,
                                double *DragZ,
                                unsigned int* neighborX,
                                unsigned int* neighborY,
                                unsigned int* neighborZ,
                                unsigned long long numberOfLBnodes, 
                                bool isEvenTimestep,
                                unsigned int numberOfThreads);

void CalcCPtop27(real* DD, 
                            int* cpIndex, 
                            int nonCp, 
                            double *cpPress,
                            unsigned int* neighborX,
                            unsigned int* neighborY,
                            unsigned int* neighborZ,
                            unsigned long long numberOfLBnodes, 
                            bool isEvenTimestep,
                            unsigned int numberOfThreads);

void CalcCPbottom27(real* DD, 
                               int* cpIndex, 
                               int nonCp, 
                               double *cpPress,
                               unsigned int* neighborX,
                               unsigned int* neighborY,
                               unsigned int* neighborZ,
                               unsigned long long numberOfLBnodes, 
                               bool isEvenTimestep,
                               unsigned int numberOfThreads);

void GetSendFsPreDev27(real* DD,
                                  real* bufferFs,
                                  int* sendIndex,
                                  int buffmax,
                                  unsigned int* neighborX,
                                  unsigned int* neighborY,
                                  unsigned int* neighborZ,
                                  unsigned long long numberOfLBnodes, 
                                  bool isEvenTimestep,
                                  unsigned int numberOfThreads, 
                                  cudaStream_t stream = CU_STREAM_LEGACY);

void GetSendFsPostDev27(real* DD,
                                   real* bufferFs,
                                   int* sendIndex,
                                   int buffmax,
                                   unsigned int* neighborX,
                                   unsigned int* neighborY,
                                   unsigned int* neighborZ,
                                   unsigned long long numberOfLBnodes, 
                                   bool isEvenTimestep,
                                   unsigned int numberOfThreads, 
                                   cudaStream_t stream = CU_STREAM_LEGACY);

void SetRecvFsPreDev27(real* DD,
                                  real* bufferFs,
                                  int* recvIndex,
                                  int buffmax,
                                  unsigned int* neighborX,
                                  unsigned int* neighborY,
                                  unsigned int* neighborZ,
                                  unsigned long long numberOfLBnodes, 
                                  bool isEvenTimestep, unsigned int numberOfThreads, 
                                  cudaStream_t stream = CU_STREAM_LEGACY);

void SetRecvFsPostDev27(real* DD,
                                   real* bufferFs,
                                   int* recvIndex,
                                   int buffmax,
                                   unsigned int* neighborX,
                                   unsigned int* neighborY,
                                   unsigned int* neighborZ,
                                   unsigned long long numberOfLBnodes, 
                                   bool isEvenTimestep,
                                   unsigned int numberOfThreads,
                                   cudaStream_t stream = CU_STREAM_LEGACY);

void getSendGsDevF3(
    real* G6,
    real* bufferGs,
    int* sendIndex,
    int buffmax,
    unsigned int* neighborX,
    unsigned int* neighborY,
    unsigned int* neighborZ,
    unsigned long long numberOfLBnodes,
    bool isEvenTimestep,
    unsigned int numberOfThreads);

void setRecvGsDevF3(
    real* G6,
    real* bufferGs,
    int* recvIndex,
    int buffmax,
    unsigned int* neighborX,
    unsigned int* neighborY,
    unsigned int* neighborZ,
    unsigned long long numberOfLBnodes,
    bool isEvenTimestep,
    unsigned int numberOfThreads);

void WallFuncDev27(unsigned int numberOfThreads,
                              real* vx,
                              real* vy,
                              real* vz,
                              real* DD, 
                              int* k_Q, 
                              real* QQ,
                              unsigned int numberOfBCnodes, 
                              real om1, 
                              unsigned int* neighborX,
                              unsigned int* neighborY,
                              unsigned int* neighborZ,
                              unsigned long long numberOfLBnodes, 
                              bool isEvenTimestep);

void SetOutputWallVelocitySP27(unsigned int numberOfThreads,
                                          real* vxD,
                                          real* vyD,
                                          real* vzD,
                                          real* vxWall,
                                          real* vyWall,
                                          real* vzWall,
                                          int numberOfWallNodes, 
                                          int* kWallNodes, 
                                          real* rhoD,
                                          real* pressD,
                                          unsigned int* geoD,
                                          unsigned int* neighborX,
                                          unsigned int* neighborY,
                                          unsigned int* neighborZ,
                                          unsigned long long numberOfLBnodes,
                                          real* DD,
                                          bool isEvenTimestep);

void CalcTurbulenceIntensityDevice(
   real* vxx,
   real* vyy,
   real* vzz,
   real* vxy,
   real* vxz,
   real* vyz,
   real* vx_mean,
   real* vy_mean,
   real* vz_mean,
   real* DD, 
   uint *typeOfGridNode, 
   unsigned int* neighborX,
   unsigned int* neighborY,
   unsigned int* neighborZ,
   unsigned long long numberOfLBnodes, 
   bool isEvenTimestep,
   uint numberOfThreads);

#endif
