//=======================================================================================
// ____          ____    __    ______     __________   __      __       __        __
// \    \       |    |  |  |  |   _   \  |___    ___| |  |    |  |     /  \      |  |
//  \    \      |    |  |  |  |  |_)   |     |  |     |  |    |  |    /    \     |  |
//   \    \     |    |  |  |  |   _   /      |  |     |  |    |  |   /  /\  \    |  |
//    \    \    |    |  |  |  |  | \  \      |  |     |   \__/   |  /  ____  \   |  |____
//     \    \   |    |  |__|  |__|  \__\     |__|      \________/  /__/    \__\  |_______|
//      \    \  |    |   ________________________________________________________________
//       \    \ |    |  |  ______________________________________________________________|
//        \    \|    |  |  |         __          __     __     __     ______      _______
//         \         |  |  |_____   |  |        |  |   |  |   |  |   |   _  \    /  _____)
//          \        |  |   _____|  |  |        |  |   |  |   |  |   |  | \  \   \_______
//           \       |  |  |        |  |_____   |   \_/   |   |  |   |  |_/  /    _____  |
//            \ _____|  |__|        |________|   \_______/    |__|   |______/    (_______/
//
//  This file is part of VirtualFluids. VirtualFluids is free software: you can
//  redistribute it and/or modify it under the terms of the GNU General Public
//  License as published by the Free Software Foundation, either version 3 of
//  the License, or (at your option) any later version.
//
//  VirtualFluids is distributed in the hope that it will be useful, but WITHOUT
//  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
//  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
//  for more details.
//
//  You should have received a copy of the GNU General Public License along
//  with VirtualFluids (see COPYING.txt). If not, see <http://www.gnu.org/licenses/>.
//
//! \author Martin Schoenherr
//=======================================================================================
#ifndef GPU_KERNELS_H
#define GPU_KERNELS_H

#include "LBM/LB.h"

__global__ void LBCalcMac27( real* vxD,
                                        real* vyD,
                                        real* vzD,
                                        real* rhoD,
                                        unsigned int* neighborX,
                                        unsigned int* neighborY,
                                        unsigned int* neighborZ,
                                        unsigned int* geoD,
                                        unsigned long long numberOfLBnodes,
                                        real* DD,
                                        bool isEvenTimestep);

__global__ void LBCalcMacSP27( real* vxD,
                                          real* vyD,
                                          real* vzD,
                                          real* rhoD,
                                          real* pressD,
                                          unsigned int* geoD,
                                          unsigned int* neighborX,
                                          unsigned int* neighborY,
                                          unsigned int* neighborZ,
                                          unsigned long long numberOfLBnodes,
                                          real* DD,
                                          bool isEvenTimestep);

__global__ void LBCalcMacCompSP27( real* vxD,
                                              real* vyD,
                                              real* vzD,
                                              real* rhoD,
                                              real* pressD,
                                              unsigned int* geoD,
                                              unsigned int* neighborX,
                                              unsigned int* neighborY,
                                              unsigned int* neighborZ,
                                              unsigned long long numberOfLBnodes,
                                              real* DD,
                                              bool isEvenTimestep);

__global__ void CalcConc7( real* Conc,
                                          unsigned int* geoD,
                                          unsigned int* neighborX,
                                          unsigned int* neighborY,
                                          unsigned int* neighborZ,
                                          unsigned long long numberOfLBnodes,
                                          real* DD7,
                                          bool isEvenTimestep);

__global__ void GetPlaneConc7(real* Conc,
                                            int* kPC,
                                            unsigned int numberOfPointskPC,
                                            unsigned int* geoD,
                                            unsigned int* neighborX,
                                            unsigned int* neighborY,
                                            unsigned int* neighborZ,
                                            unsigned long long numberOfLBnodes,
                                            real* DD7,
                                            bool isEvenTimestep);

__global__ void GetPlaneConc27(real* Conc,
                                             int* kPC,
                                             unsigned int numberOfPointskPC,
                                             unsigned int* geoD,
                                             unsigned int* neighborX,
                                             unsigned int* neighborY,
                                             unsigned int* neighborZ,
                                             unsigned long long numberOfLBnodes,
                                             real* DD27,
                                             bool isEvenTimestep);

__global__ void CalcConc27(real* Conc,
                                          unsigned int* geoD,
                                          unsigned int* neighborX,
                                          unsigned int* neighborY,
                                          unsigned int* neighborZ,
                                          unsigned long long numberOfLBnodes,
                                          real* DD27,
                                          bool isEvenTimestep);

__global__ void LBCalcMedSP27( real* vxD,
                                          real* vyD,
                                          real* vzD,
                                          real* rhoD,
                                          real* pressD,
                                          unsigned int* geoD,
                                          unsigned int* neighborX,
                                          unsigned int* neighborY,
                                          unsigned int* neighborZ,
                                          unsigned long long numberOfLBnodes,
                                          real* DD,
                                          bool isEvenTimestep);

__global__ void LBCalcMedCompSP27( real* vxD,
                                              real* vyD,
                                              real* vzD,
                                              real* rhoD,
                                              real* pressD,
                                              unsigned int* geoD,
                                              unsigned int* neighborX,
                                              unsigned int* neighborY,
                                              unsigned int* neighborZ,
                                              unsigned long long numberOfLBnodes,
                                              real* DD,
                                              bool isEvenTimestep);

__global__ void LBCalcMedCompAD27(
    real* vxD,
    real* vyD,
    real* vzD,
    real* rhoD,
    real* pressD,
    real* concD,
    unsigned int* geoD,
    unsigned int* neighborX,
    unsigned int* neighborY,
    unsigned int* neighborZ,
    unsigned long long numberOfLBnodes,
    real* DD,
    real* DD_AD,
    bool isEvenTimestep);

__global__ void LBCalcMacMedSP27( real* vxD,
                                             real* vyD,
                                             real* vzD,
                                             real* rhoD,
                                             real* pressD,
                                             unsigned int* geoD,
                                             unsigned int* neighborX,
                                             unsigned int* neighborY,
                                             unsigned int* neighborZ,
                                             unsigned int tdiff,
                                             unsigned long long numberOfLBnodes,
                                             bool isEvenTimestep);

__global__ void LBResetMeanValuesSP27(
    real* vxD,
    real* vyD,
    real* vzD,
    real* rhoD,
    real* pressD,
    unsigned long long numberOfLBnodes,
    bool isEvenTimestep);

__global__ void LBResetMeanValuesAD27(
    real* vxD,
    real* vyD,
    real* vzD,
    real* rhoD,
    real* pressD,
    real* concD,
    unsigned long long numberOfLBnodes,
    bool isEvenTimestep);

__global__ void LBCalc2ndMomentsIncompSP27(  real* kxyFromfcNEQ,
                                                        real* kyzFromfcNEQ,
                                                        real* kxzFromfcNEQ,
                                                        real* kxxMyyFromfcNEQ,
                                                        real* kxxMzzFromfcNEQ,
                                                        unsigned int* geoD,
                                                        unsigned int* neighborX,
                                                        unsigned int* neighborY,
                                                        unsigned int* neighborZ,
                                                        unsigned long long numberOfLBnodes,
                                                        real* DD,
                                                        bool isEvenTimestep);

__global__ void LBCalc2ndMomentsCompSP27(real* kxyFromfcNEQ,
                                                    real* kyzFromfcNEQ,
                                                    real* kxzFromfcNEQ,
                                                    real* kxxMyyFromfcNEQ,
                                                    real* kxxMzzFromfcNEQ,
                                                    unsigned int* geoD,
                                                    unsigned int* neighborX,
                                                    unsigned int* neighborY,
                                                    unsigned int* neighborZ,
                                                    unsigned long long numberOfLBnodes,
                                                    real* DD,
                                                    bool isEvenTimestep);

__global__ void LBCalc3rdMomentsIncompSP27(  real* CUMbbb,
                                                        real* CUMabc,
                                                        real* CUMbac,
                                                        real* CUMbca,
                                                        real* CUMcba,
                                                        real* CUMacb,
                                                        real* CUMcab,
                                                        unsigned int* bcMatD,
                                                        unsigned int* neighborX,
                                                        unsigned int* neighborY,
                                                        unsigned int* neighborZ,
                                                        real* DDStart,
                                                        unsigned long long numberOfLBnodes,
                                                        bool EvenOrOdd);

__global__ void LBCalc3rdMomentsCompSP27(real* CUMbbb,
                                                    real* CUMabc,
                                                    real* CUMbac,
                                                    real* CUMbca,
                                                    real* CUMcba,
                                                    real* CUMacb,
                                                    real* CUMcab,
                                                    unsigned int* bcMatD,
                                                    unsigned int* neighborX,
                                                    unsigned int* neighborY,
                                                    unsigned int* neighborZ,
                                                    real* DDStart,
                                                    unsigned long long numberOfLBnodes,
                                                    bool EvenOrOdd);

__global__ void LBCalcHigherMomentsIncompSP27(   real* CUMcbb,
                                                            real* CUMbcb,
                                                            real* CUMbbc,
                                                            real* CUMcca,
                                                            real* CUMcac,
                                                            real* CUMacc,
                                                            real* CUMbcc,
                                                            real* CUMcbc,
                                                            real* CUMccb,
                                                            real* CUMccc,
                                                            unsigned int* bcMatD,
                                                            unsigned int* neighborX,
                                                            unsigned int* neighborY,
                                                            unsigned int* neighborZ,
                                                            real* DDStart,
                                                            unsigned long long numberOfLBnodes,
                                                            bool EvenOrOdd);

__global__ void LBCalcHigherMomentsCompSP27( real* CUMcbb,
                                                        real* CUMbcb,
                                                        real* CUMbbc,
                                                        real* CUMcca,
                                                        real* CUMcac,
                                                        real* CUMacc,
                                                        real* CUMbcc,
                                                        real* CUMcbc,
                                                        real* CUMccb,
                                                        real* CUMccc,
                                                        unsigned int* bcMatD,
                                                        unsigned int* neighborX,
                                                        unsigned int* neighborY,
                                                        unsigned int* neighborZ,
                                                        real* DDStart,
                                                        unsigned long long numberOfLBnodes,
                                                        bool EvenOrOdd);

__global__ void LBCalcMeasurePoints(real* vxMP,
                                               real* vyMP,
                                               real* vzMP,
                                               real* rhoMP,
                                               unsigned int* kMP,
                                               unsigned int numberOfPointskMP,
                                               unsigned int MPClockCycle,
                                               unsigned int t,
                                               unsigned int* geoD,
                                               unsigned int* neighborX,
                                               unsigned int* neighborY,
                                               unsigned int* neighborZ,
                                               unsigned long long numberOfLBnodes,
                                               real* DD,
                                               bool isEvenTimestep);

//Advection / Diffusion BCs
__global__ void QAD7( real* DD,
                                 real* DD7,
                                 real* temp,
                                 real diffusivity,
                                 int* k_Q,
                                 real* QQ,
                                 unsigned int numberOfBCnodes,
                                 real om1,
                                 unsigned int* neighborX,
                                 unsigned int* neighborY,
                                 unsigned int* neighborZ,
                                 unsigned long long numberOfLBnodes,
                                 bool isEvenTimestep);

//////////////////////////////////////////////////////////////////////////
//! \brief \ref AD_SlipVelDeviceComp : device function for the slip-AD boundary condition
__global__ void AD_SlipVelDeviceComp(
    real * normalX,
    real * normalY,
    real * normalZ,
    real * distributions,
    real * distributionsAD,
    int* QindexArray,
    real * Qarrays,
    uint numberOfBCnodes,
    real omegaDiffusivity,
    uint * neighborX,
    uint * neighborY,
    uint * neighborZ,
    unsigned long long numberOfLBnodes,
    bool isEvenTimestep);

__global__ void QADDirichlet27(   real* DD,
                                             real* DD27,
                                             real* temp,
                                             real diffusivity,
                                             int* k_Q,
                                             real* QQ,
                                             unsigned int numberOfBCnodes,
                                             real om1,
                                             unsigned int* neighborX,
                                             unsigned int* neighborY,
                                             unsigned int* neighborZ,
                                             unsigned long long numberOfLBnodes,
                                             bool isEvenTimestep);

__global__ void QADBB27(  real* DD,
                                     real* DD27,
                                     real* temp,
                                     real diffusivity,
                                     int* k_Q,
                                     real* QQ,
                                     unsigned int numberOfBCnodes,
                                     real om1,
                                     unsigned int* neighborX,
                                     unsigned int* neighborY,
                                     unsigned int* neighborZ,
                                     unsigned long long numberOfLBnodes,
                                     bool isEvenTimestep);

__global__ void QADVel7( real* DD,
                                    real* DD7,
                                    real* temp,
                                    real* velo,
                                    real diffusivity,
                                    int* k_Q,
                                    real* QQ,
                                    unsigned int numberOfBCnodes,
                                    real om1,
                                    unsigned int* neighborX,
                                    unsigned int* neighborY,
                                    unsigned int* neighborZ,
                                    unsigned long long numberOfLBnodes,
                                    bool isEvenTimestep);

__global__ void QADVel27(real* DD,
                                    real* DD27,
                                    real* temp,
                                    real* velo,
                                    real diffusivity,
                                    int* k_Q,
                                    real* QQ,
                                    unsigned int numberOfBCnodes,
                                    real om1,
                                    unsigned int* neighborX,
                                    unsigned int* neighborY,
                                    unsigned int* neighborZ,
                                    unsigned long long numberOfLBnodes,
                                    bool isEvenTimestep);

__global__ void QADPress7(  real* DD,
                                       real* DD7,
                                       real* temp,
                                       real* velo,
                                       real diffusivity,
                                       int* k_Q,
                                       real* QQ,
                                       unsigned int numberOfBCnodes,
                                       real om1,
                                       unsigned int* neighborX,
                                       unsigned int* neighborY,
                                       unsigned int* neighborZ,
                                       unsigned long long numberOfLBnodes,
                                       bool isEvenTimestep);

__global__ void QADPress27( real* DD,
                                       real* DD27,
                                       real* temp,
                                       real* velo,
                                       real diffusivity,
                                       int* k_Q,
                                       real* QQ,
                                       unsigned int numberOfBCnodes,
                                       real om1,
                                       unsigned int* neighborX,
                                       unsigned int* neighborY,
                                       unsigned int* neighborZ,
                                       unsigned long long numberOfLBnodes,
                                       bool isEvenTimestep);

__global__ void QADPressNEQNeighbor27(
                                                 real* DD,
                                                 real* DD27,
                                                 int* k_Q,
                                                 int* k_N,
                                                 int numberOfBCnodes,
                                                 unsigned int* neighborX,
                                                 unsigned int* neighborY,
                                                 unsigned int* neighborZ,
                                                 unsigned long long numberOfLBnodes,
                                                 bool isEvenTimestep
                                                );

__global__ void QNoSlipADincomp7( real* DD,
                                             real* DD7,
                                             real* temp,
                                             real diffusivity,
                                             int* k_Q,
                                             real* QQ,
                                             unsigned int numberOfBCnodes,
                                             real om1,
                                             unsigned int* neighborX,
                                             unsigned int* neighborY,
                                             unsigned int* neighborZ,
                                             unsigned long long numberOfLBnodes,
                                             bool isEvenTimestep);

__global__ void QNoSlipADincomp27( real* DD,
                                             real* DD27,
                                             real* temp,
                                             real diffusivity,
                                             int* k_Q,
                                             real* QQ,
                                             unsigned int numberOfBCnodes,
                                             real om1,
                                             unsigned int* neighborX,
                                             unsigned int* neighborY,
                                             unsigned int* neighborZ,
                                             unsigned long long numberOfLBnodes,
                                             bool isEvenTimestep);

__global__ void QADVeloIncomp7(  real* DD,
                                            real* DD7,
                                            real* temp,
                                            real* velo,
                                            real diffusivity,
                                            int* k_Q,
                                            real* QQ,
                                            unsigned int numberOfBCnodes,
                                            real om1,
                                            unsigned int* neighborX,
                                            unsigned int* neighborY,
                                            unsigned int* neighborZ,
                                            unsigned long long numberOfLBnodes,
                                            bool isEvenTimestep);

__global__ void QADVeloIncomp27( real* DD,
                                            real* DD27,
                                            real* temp,
                                            real* velo,
                                            real diffusivity,
                                            int* k_Q,
                                            real* QQ,
                                            unsigned int numberOfBCnodes,
                                            real om1,
                                            unsigned int* neighborX,
                                            unsigned int* neighborY,
                                            unsigned int* neighborZ,
                                            unsigned long long numberOfLBnodes,
                                            bool isEvenTimestep);

__global__ void QADPressIncomp7(real* DD,
                                           real* DD7,
                                           real* temp,
                                           real* velo,
                                           real diffusivity,
                                           int* k_Q,
                                           real* QQ,
                                           unsigned int numberOfBCnodes,
                                           real om1,
                                           unsigned int* neighborX,
                                           unsigned int* neighborY,
                                           unsigned int* neighborZ,
                                           unsigned long long numberOfLBnodes,
                                           bool isEvenTimestep);

__global__ void QADPressIncomp27(   real* DD,
                                               real* DD27,
                                               real* temp,
                                               real* velo,
                                               real diffusivity,
                                               int* k_Q,
                                               real* QQ,
                                               unsigned int numberOfBCnodes,
                                               real om1,
                                               unsigned int* neighborX,
                                               unsigned int* neighborY,
                                               unsigned int* neighborZ,
                                               unsigned long long numberOfLBnodes,
                                               bool isEvenTimestep);



//coarse to fine
__global__ void scaleCF27(real* DC,
                                     real* DF,
                                    unsigned int* neighborCX,
                                    unsigned int* neighborCY,
                                    unsigned int* neighborCZ,
                                    unsigned int* neighborFX,
                                    unsigned int* neighborFY,
                                    unsigned int* neighborFZ,
                                               unsigned long long numberOfLBnodesC,
                                               unsigned long long numberOfLBnodesF,
                                               bool isEvenTimestep,
                                     unsigned int* posCSWB,
                                     unsigned int* posFSWB,
                                     unsigned int kCF,
                                               real omCoarse,
                                               real omFine,
                                               real nu,
                                               unsigned int nxC,
                                               unsigned int nyC,
                                               unsigned int nxF,
                                               unsigned int nyF);

__global__ void scaleCFEff27(real* DC,
                                        real* DF,
                                        unsigned int* neighborCX,
                                        unsigned int* neighborCY,
                                        unsigned int* neighborCZ,
                                        unsigned int* neighborFX,
                                        unsigned int* neighborFY,
                                        unsigned int* neighborFZ,
                                                 unsigned long long numberOfLBnodesC,
                                                 unsigned long long numberOfLBnodesF,
                                                 bool isEvenTimestep,
                                        unsigned int* posCSWB,
                                        unsigned int* posFSWB,
                                        unsigned int kCF,
                                                 real omCoarse,
                                                 real omFine,
                                                 real nu,
                                                 unsigned int nxC,
                                                 unsigned int nyC,
                                                 unsigned int nxF,
                                        unsigned int nyF,
                                        ICellNeigh neighborCoarseToFine);

__global__ void scaleCFLast27( real* DC,
                                          real* DF,
                                          unsigned int* neighborCX,
                                          unsigned int* neighborCY,
                                          unsigned int* neighborCZ,
                                          unsigned int* neighborFX,
                                          unsigned int* neighborFY,
                                          unsigned int* neighborFZ,
                                          unsigned long long numberOfLBnodesC,
                                          unsigned long long numberOfLBnodesF,
                                          bool isEvenTimestep,
                                          unsigned int* posCSWB,
                                          unsigned int* posFSWB,
                                          unsigned int kCF,
                                          real omCoarse,
                                          real omFine,
                                          real nu,
                                          unsigned int nxC,
                                          unsigned int nyC,
                                          unsigned int nxF,
                                          unsigned int nyF,
                                          ICellNeigh neighborCoarseToFine);

__global__ void scaleCFpress27(real* DC,
                                          real* DF,
                                          unsigned int* neighborCX,
                                          unsigned int* neighborCY,
                                          unsigned int* neighborCZ,
                                          unsigned int* neighborFX,
                                          unsigned int* neighborFY,
                                          unsigned int* neighborFZ,
                                          unsigned long long numberOfLBnodesC,
                                          unsigned long long numberOfLBnodesF,
                                          bool isEvenTimestep,
                                          unsigned int* posCSWB,
                                          unsigned int* posFSWB,
                                          unsigned int kCF,
                                          real omCoarse,
                                          real omFine,
                                          real nu,
                                          unsigned int nxC,
                                          unsigned int nyC,
                                          unsigned int nxF,
                                          unsigned int nyF,
                                          ICellNeigh neighborCoarseToFine);

__global__ void scaleCF_Fix_27(real* DC,
                                          real* DF,
                                          unsigned int* neighborCX,
                                          unsigned int* neighborCY,
                                          unsigned int* neighborCZ,
                                          unsigned int* neighborFX,
                                          unsigned int* neighborFY,
                                          unsigned int* neighborFZ,
                                          unsigned long long numberOfLBnodesC,
                                          unsigned long long numberOfLBnodesF,
                                          bool isEvenTimestep,
                                          unsigned int* posCSWB,
                                          unsigned int* posFSWB,
                                          unsigned int kCF,
                                          real omCoarse,
                                          real omFine,
                                          real nu,
                                          unsigned int nxC,
                                          unsigned int nyC,
                                          unsigned int nxF,
                                          unsigned int nyF,
                                          ICellNeigh neighborCoarseToFine);

__global__ void scaleCF_Fix_comp_27(   real* DC,
                                                  real* DF,
                                                  unsigned int* neighborCX,
                                                  unsigned int* neighborCY,
                                                  unsigned int* neighborCZ,
                                                  unsigned int* neighborFX,
                                                  unsigned int* neighborFY,
                                                  unsigned int* neighborFZ,
                                                  unsigned long long numberOfLBnodesC,
                                                  unsigned long long numberOfLBnodesF,
                                                  bool isEvenTimestep,
                                                  unsigned int* posCSWB,
                                                  unsigned int* posFSWB,
                                                  unsigned int kCF,
                                                  real omCoarse,
                                                  real omFine,
                                                  real nu,
                                                  unsigned int nxC,
                                                  unsigned int nyC,
                                                  unsigned int nxF,
                                                  unsigned int nyF,
                                                  ICellNeigh neighborCoarseToFine);

__global__ void scaleCF_0817_comp_27(  real* DC,
                                                  real* DF,
                                                  unsigned int* neighborCX,
                                                  unsigned int* neighborCY,
                                                  unsigned int* neighborCZ,
                                                  unsigned int* neighborFX,
                                                  unsigned int* neighborFY,
                                                  unsigned int* neighborFZ,
                                                  unsigned long long numberOfLBnodesC,
                                                  unsigned long long numberOfLBnodesF,
                                                  bool isEvenTimestep,
                                                  unsigned int* posCSWB,
                                                  unsigned int* posFSWB,
                                                  unsigned int kCF,
                                                  real omCoarse,
                                                  real omFine,
                                                  real nu,
                                                  unsigned int nxC,
                                                  unsigned int nyC,
                                                  unsigned int nxF,
                                                  unsigned int nyF,
                                                  ICellNeigh neighborCoarseToFine);

__global__ void scaleCF_comp_D3Q27F3_2018( real* DC,
                                                      real* DF,
                                                      real* G6,
                                                      unsigned int* neighborCX,
                                                      unsigned int* neighborCY,
                                                      unsigned int* neighborCZ,
                                                      unsigned int* neighborFX,
                                                      unsigned int* neighborFY,
                                                      unsigned int* neighborFZ,
                                                      unsigned long long numberOfLBnodesC,
                                                      unsigned long long numberOfLBnodesF,
                                                      bool isEvenTimestep,
                                                      unsigned int* posCSWB,
                                                      unsigned int* posFSWB,
                                                      unsigned int kCF,
                                                      real omCoarse,
                                                      real omFine,
                                                      real nu,
                                                      unsigned int nxC,
                                                      unsigned int nyC,
                                                      unsigned int nxF,
                                                      unsigned int nyF,
                                                      ICellNeigh neighborCoarseToFine);

__global__ void scaleCF_comp_D3Q27F3( real* DC,
                                                 real* DF,
                                                 real* G6,
                                                 unsigned int* neighborCX,
                                                 unsigned int* neighborCY,
                                                 unsigned int* neighborCZ,
                                                 unsigned int* neighborFX,
                                                 unsigned int* neighborFY,
                                                 unsigned int* neighborFZ,
                                                 unsigned long long numberOfLBnodesC,
                                                 unsigned long long numberOfLBnodesF,
                                                 bool isEvenTimestep,
                                                 unsigned int* posCSWB,
                                                 unsigned int* posFSWB,
                                                 unsigned int kCF,
                                                 real omCoarse,
                                                 real omFine,
                                                 real nu,
                                                 unsigned int nxC,
                                                 unsigned int nyC,
                                                 unsigned int nxF,
                                                 unsigned int nyF,
                                                 ICellNeigh neighborCoarseToFine);


__global__ void scaleCF_staggered_time_comp_27(real* DC,
                                                          real* DF,
                                                          unsigned int* neighborCX,
                                                          unsigned int* neighborCY,
                                                          unsigned int* neighborCZ,
                                                          unsigned int* neighborFX,
                                                          unsigned int* neighborFY,
                                                          unsigned int* neighborFZ,
                                                          unsigned long long numberOfLBnodesC,
                                                          unsigned long long numberOfLBnodesF,
                                                          bool isEvenTimestep,
                                                          unsigned int* posCSWB,
                                                          unsigned int* posFSWB,
                                                          unsigned int kCF,
                                                          real omCoarse,
                                                          real omFine,
                                                          real nu,
                                                          unsigned int nxC,
                                                          unsigned int nyC,
                                                          unsigned int nxF,
                                                          unsigned int nyF,
                                                          ICellNeigh neighborCoarseToFine);

__global__ void scaleCF_RhoSq_comp_27( real* DC,
                                                  real* DF,
                                                  unsigned int* neighborCX,
                                                  unsigned int* neighborCY,
                                                  unsigned int* neighborCZ,
                                                  unsigned int* neighborFX,
                                                  unsigned int* neighborFY,
                                                  unsigned int* neighborFZ,
                                                  unsigned long long numberOfLBnodesC,
                                                  unsigned long long numberOfLBnodesF,
                                                  bool isEvenTimestep,
                                                  unsigned int* posCSWB,
                                                  unsigned int* posFSWB,
                                                  unsigned int kCF,
                                                  real omCoarse,
                                                  real omFine,
                                                  real nu,
                                                  unsigned int nxC,
                                                  unsigned int nyC,
                                                  unsigned int nxF,
                                                  unsigned int nyF,
                                                  ICellNeigh neighborCoarseToFine);

template<bool hasTurbulentViscosity> __global__ void scaleCF_compressible(
    real* distributionsCoarse,
    real* distributionsFine,
    unsigned int* neighborXcoarse,
    unsigned int* neighborYcoarse,
    unsigned int* neighborZcoarse,
    unsigned int* neighborXfine,
    unsigned int* neighborYfine,
    unsigned int* neighborZfine,
    unsigned long long numberOfLBnodesCoarse,
    unsigned long long numberOfLBnodesFine,
    bool isEvenTimestep,
    unsigned int* indicesCoarseMMM,
    unsigned int* indicesFineMMM,
    unsigned int numberOfInterfaceNodes,
    real omegaCoarse,
    real omegaFine,
    real* turbulentViscosityCoarse,
    real* turbulentViscosityFine,
    ICellNeigh offsetCF);

__global__ void scaleCF_RhoSq_3rdMom_comp_27(real* DC,
                                                        real* DF,
                                                        unsigned int* neighborCX,
                                                        unsigned int* neighborCY,
                                                        unsigned int* neighborCZ,
                                                        unsigned int* neighborFX,
                                                        unsigned int* neighborFY,
                                                        unsigned int* neighborFZ,
                                                        unsigned long long numberOfLBnodesC,
                                                        unsigned long long numberOfLBnodesF,
                                                        bool isEvenTimestep,
                                                        unsigned int* posCSWB,
                                                        unsigned int* posFSWB,
                                                        unsigned int kCF,
                                                        real omCoarse,
                                                        real omFine,
                                                        real nu,
                                                        unsigned int nxC,
                                                        unsigned int nyC,
                                                        unsigned int nxF,
                                                        unsigned int nyF,
                                                        ICellNeigh neighborCoarseToFine);

__global__ void scaleCF_AA2016_comp_27(real* DC,
                                                  real* DF,
                                                  unsigned int* neighborCX,
                                                  unsigned int* neighborCY,
                                                  unsigned int* neighborCZ,
                                                  unsigned int* neighborFX,
                                                  unsigned int* neighborFY,
                                                  unsigned int* neighborFZ,
                                                  unsigned long long numberOfLBnodesC,
                                                  unsigned long long numberOfLBnodesF,
                                                  bool isEvenTimestep,
                                                  unsigned int* posCSWB,
                                                  unsigned int* posFSWB,
                                                  unsigned int kCF,
                                                  real omCoarse,
                                                  real omFine,
                                                  real nu,
                                                  unsigned int nxC,
                                                  unsigned int nyC,
                                                  unsigned int nxF,
                                                  unsigned int nyF,
                                                  ICellNeigh neighborCoarseToFine);

__global__ void scaleCF_NSPress_27(real* DC,
                                              real* DF,
                                              unsigned int* neighborCX,
                                              unsigned int* neighborCY,
                                              unsigned int* neighborCZ,
                                              unsigned int* neighborFX,
                                              unsigned int* neighborFY,
                                              unsigned int* neighborFZ,
                                              unsigned long long numberOfLBnodesC,
                                              unsigned long long numberOfLBnodesF,
                                              bool isEvenTimestep,
                                              unsigned int* posCSWB,
                                              unsigned int* posFSWB,
                                              unsigned int kCF,
                                              real omCoarse,
                                              real omFine,
                                              real nu,
                                              unsigned int nxC,
                                              unsigned int nyC,
                                              unsigned int nxF,
                                              unsigned int nyF,
                                              ICellNeigh neighborCoarseToFine);

__global__ void scaleCFThSMG7( real* DC,
                                          real* DF,
                                          real* DD7C,
                                          real* DD7F,
                                          unsigned int* neighborCX,
                                          unsigned int* neighborCY,
                                          unsigned int* neighborCZ,
                                          unsigned int* neighborFX,
                                          unsigned int* neighborFY,
                                          unsigned int* neighborFZ,
                                          unsigned long long numberOfLBnodesC,
                                          unsigned long long numberOfLBnodesF,
                                          bool isEvenTimestep,
                                          unsigned int* posCSWB,
                                          unsigned int* posFSWB,
                                          unsigned int kCF,
                                          real nu,
                                          real diffusivity_fine,
                                          ICellNeigh neighborCoarseToFine);

__global__ void scaleCFThS7(real* DC,
                                       real* DF,
                                       real* DD7C,
                                       real* DD7F,
                                       unsigned int* neighborCX,
                                       unsigned int* neighborCY,
                                       unsigned int* neighborCZ,
                                       unsigned int* neighborFX,
                                       unsigned int* neighborFY,
                                       unsigned int* neighborFZ,
                                       unsigned long long numberOfLBnodesC,
                                       unsigned long long numberOfLBnodesF,
                                       bool isEvenTimestep,
                                       unsigned int* posCSWB,
                                       unsigned int* posFSWB,
                                       unsigned int kCF,
                                       real nu,
                                       real diffusivity_fine);

__global__ void scaleCFThS27(real* DC,
                                        real* DF,
                                        real* DD27C,
                                        real* DD27F,
                                        unsigned int* neighborCX,
                                        unsigned int* neighborCY,
                                        unsigned int* neighborCZ,
                                        unsigned int* neighborFX,
                                        unsigned int* neighborFY,
                                        unsigned int* neighborFZ,
                                        unsigned long long numberOfLBnodesC,
                                        unsigned long long numberOfLBnodesF,
                                        bool isEvenTimestep,
                                        unsigned int* posCSWB,
                                        unsigned int* posFSWB,
                                        unsigned int kCF,
                                        real nu,
                                        real diffusivity_fine,
                                        ICellNeigh neighborCoarseToFine);

//fine to coarse
__global__ void scaleFC27(real* DC,
                                     real* DF,
                                    unsigned int* neighborCX,
                                    unsigned int* neighborCY,
                                    unsigned int* neighborCZ,
                                    unsigned int* neighborFX,
                                    unsigned int* neighborFY,
                                    unsigned int* neighborFZ,
                                               unsigned long long numberOfLBnodesC,
                                               unsigned long long numberOfLBnodesF,
                                               bool isEvenTimestep,
                                     unsigned int* posC,
                                     unsigned int* posFSWB,
                                     unsigned int kFC,
                                               real omCoarse,
                                               real omFine,
                                               real nu,
                                               unsigned int nxC,
                                               unsigned int nyC,
                                               unsigned int nxF,
                                     unsigned int nyF);

__global__ void scaleFCEff27(real* DC,
                                        real* DF,
                                        unsigned int* neighborCX,
                                        unsigned int* neighborCY,
                                        unsigned int* neighborCZ,
                                        unsigned int* neighborFX,
                                        unsigned int* neighborFY,
                                        unsigned int* neighborFZ,
                                        unsigned long long numberOfLBnodesC,
                                        unsigned long long numberOfLBnodesF,
                                        bool isEvenTimestep,
                                        unsigned int* posC,
                                        unsigned int* posFSWB,
                                        unsigned int kFC,
                                        real omCoarse,
                                        real omFine,
                                        real nu,
                                        unsigned int nxC,
                                        unsigned int nyC,
                                        unsigned int nxF,
                                        unsigned int nyF,
                                        ICellNeigh neighborFineToCoarse);

__global__ void scaleFCLast27( real* DC,
                                          real* DF,
                                          unsigned int* neighborCX,
                                          unsigned int* neighborCY,
                                          unsigned int* neighborCZ,
                                          unsigned int* neighborFX,
                                          unsigned int* neighborFY,
                                          unsigned int* neighborFZ,
                                          unsigned long long numberOfLBnodesC,
                                          unsigned long long numberOfLBnodesF,
                                          bool isEvenTimestep,
                                          unsigned int* posC,
                                          unsigned int* posFSWB,
                                          unsigned int kFC,
                                          real omCoarse,
                                          real omFine,
                                          real nu,
                                          unsigned int nxC,
                                          unsigned int nyC,
                                          unsigned int nxF,
                                          unsigned int nyF,
                                          ICellNeigh neighborFineToCoarse);

__global__ void scaleFCpress27( real* DC,
                                          real* DF,
                                          unsigned int* neighborCX,
                                          unsigned int* neighborCY,
                                          unsigned int* neighborCZ,
                                          unsigned int* neighborFX,
                                          unsigned int* neighborFY,
                                          unsigned int* neighborFZ,
                                          unsigned long long numberOfLBnodesC,
                                          unsigned long long numberOfLBnodesF,
                                          bool isEvenTimestep,
                                          unsigned int* posC,
                                          unsigned int* posFSWB,
                                          unsigned int kFC,
                                          real omCoarse,
                                          real omFine,
                                          real nu,
                                          unsigned int nxC,
                                          unsigned int nyC,
                                          unsigned int nxF,
                                          unsigned int nyF,
                                          ICellNeigh neighborFineToCoarse);

__global__ void scaleFC_Fix_27( real* DC,
                                          real* DF,
                                          unsigned int* neighborCX,
                                          unsigned int* neighborCY,
                                          unsigned int* neighborCZ,
                                          unsigned int* neighborFX,
                                          unsigned int* neighborFY,
                                          unsigned int* neighborFZ,
                                          unsigned long long numberOfLBnodesC,
                                          unsigned long long numberOfLBnodesF,
                                          bool isEvenTimestep,
                                          unsigned int* posC,
                                          unsigned int* posFSWB,
                                          unsigned int kFC,
                                          real omCoarse,
                                          real omFine,
                                          real nu,
                                          unsigned int nxC,
                                          unsigned int nyC,
                                          unsigned int nxF,
                                          unsigned int nyF,
                                          ICellNeigh neighborFineToCoarse);

__global__ void scaleFC_Fix_comp_27(   real* DC,
                                                  real* DF,
                                                  unsigned int* neighborCX,
                                                  unsigned int* neighborCY,
                                                  unsigned int* neighborCZ,
                                                  unsigned int* neighborFX,
                                                  unsigned int* neighborFY,
                                                  unsigned int* neighborFZ,
                                                  unsigned long long numberOfLBnodesC,
                                                  unsigned long long numberOfLBnodesF,
                                                  bool isEvenTimestep,
                                                  unsigned int* posC,
                                                  unsigned int* posFSWB,
                                                  unsigned int kFC,
                                                  real omCoarse,
                                                  real omFine,
                                                  real nu,
                                                  unsigned int nxC,
                                                  unsigned int nyC,
                                                  unsigned int nxF,
                                                  unsigned int nyF,
                                                  ICellNeigh neighborFineToCoarse);

__global__ void scaleFC_0817_comp_27(  real* DC,
                                                  real* DF,
                                                  unsigned int* neighborCX,
                                                  unsigned int* neighborCY,
                                                  unsigned int* neighborCZ,
                                                  unsigned int* neighborFX,
                                                  unsigned int* neighborFY,
                                                  unsigned int* neighborFZ,
                                                  unsigned long long numberOfLBnodesC,
                                                  unsigned long long numberOfLBnodesF,
                                                  bool isEvenTimestep,
                                                  unsigned int* posC,
                                                  unsigned int* posFSWB,
                                                  unsigned int kFC,
                                                  real omCoarse,
                                                  real omFine,
                                                  real nu,
                                                  unsigned int nxC,
                                                  unsigned int nyC,
                                                  unsigned int nxF,
                                                  unsigned int nyF,
                                                  ICellNeigh neighborFineToCoarse);

__global__ void scaleFC_comp_D3Q27F3_2018( real* DC,
                                                      real* DF,
                                                      real* G6,
                                                      unsigned int* neighborCX,
                                                      unsigned int* neighborCY,
                                                      unsigned int* neighborCZ,
                                                      unsigned int* neighborFX,
                                                      unsigned int* neighborFY,
                                                      unsigned int* neighborFZ,
                                                      unsigned long long numberOfLBnodesC,
                                                      unsigned long long numberOfLBnodesF,
                                                      bool isEvenTimestep,
                                                      unsigned int* posC,
                                                      unsigned int* posFSWB,
                                                      unsigned int kFC,
                                                      real omCoarse,
                                                      real omFine,
                                                      real nu,
                                                      unsigned int nxC,
                                                      unsigned int nyC,
                                                      unsigned int nxF,
                                                      unsigned int nyF,
                                                      ICellNeigh neighborFineToCoarse);

__global__ void scaleFC_comp_D3Q27F3( real* DC,
                                                 real* DF,
                                                 real* G6,
                                                 unsigned int* neighborCX,
                                                 unsigned int* neighborCY,
                                                 unsigned int* neighborCZ,
                                                 unsigned int* neighborFX,
                                                 unsigned int* neighborFY,
                                                 unsigned int* neighborFZ,
                                                 unsigned long long numberOfLBnodesC,
                                                 unsigned long long numberOfLBnodesF,
                                                 bool isEvenTimestep,
                                                 unsigned int* posC,
                                                 unsigned int* posFSWB,
                                                 unsigned int kFC,
                                                 real omCoarse,
                                                 real omFine,
                                                 real nu,
                                                 unsigned int nxC,
                                                 unsigned int nyC,
                                                 unsigned int nxF,
                                                 unsigned int nyF,
                                                 ICellNeigh neighborFineToCoarse);


__global__ void scaleFC_staggered_time_comp_27(real* DC,
                                                          real* DF,
                                                          unsigned int* neighborCX,
                                                          unsigned int* neighborCY,
                                                          unsigned int* neighborCZ,
                                                          unsigned int* neighborFX,
                                                          unsigned int* neighborFY,
                                                          unsigned int* neighborFZ,
                                                          unsigned long long numberOfLBnodesC,
                                                          unsigned long long numberOfLBnodesF,
                                                          bool isEvenTimestep,
                                                          unsigned int* posC,
                                                          unsigned int* posFSWB,
                                                          unsigned int kFC,
                                                          real omCoarse,
                                                          real omFine,
                                                          real nu,
                                                          unsigned int nxC,
                                                          unsigned int nyC,
                                                          unsigned int nxF,
                                                          unsigned int nyF,
                                                          ICellNeigh neighborFineToCoarse);

__global__ void scaleFC_RhoSq_comp_27( real* DC,
                                                  real* DF,
                                                  unsigned int* neighborCX,
                                                  unsigned int* neighborCY,
                                                  unsigned int* neighborCZ,
                                                  unsigned int* neighborFX,
                                                  unsigned int* neighborFY,
                                                  unsigned int* neighborFZ,
                                                  unsigned long long numberOfLBnodesC,
                                                  unsigned long long numberOfLBnodesF,
                                                  bool isEvenTimestep,
                                                  unsigned int* posC,
                                                  unsigned int* posFSWB,
                                                  unsigned int kFC,
                                                  real omCoarse,
                                                  real omFine,
                                                  real nu,
                                                  unsigned int nxC,
                                                  unsigned int nyC,
                                                  unsigned int nxF,
                                                  unsigned int nyF,
                                                  ICellNeigh neighborFineToCoarse);

template<bool hasTurbulentViscosity> __global__ void scaleFC_compressible(
    real *distributionsCoarse,
    real *distributionsFine,
    unsigned int *neighborXcoarse,
    unsigned int *neighborYcoarse,
    unsigned int *neighborZcoarse,
    unsigned int *neighborXfine,
    unsigned int *neighborYfine,
    unsigned int *neighborZfine,
    unsigned long long numberOfLBnodesCoarse,
    unsigned long long numberOfLBnodesFine,
    bool isEvenTimestep,
    unsigned int *indicesCoarse000,
    unsigned int *indicesFineMMM,
    unsigned int numberOfInterfaceNodes,
    real omegaCoarse,
    real omegaFine,
    real* turbulentViscosityCoarse,
    real* turbulentViscosityFine,
    ICellNeigh offsetFC);

__global__ void scaleFC_RhoSq_3rdMom_comp_27(real* DC,
                                                        real* DF,
                                                        unsigned int* neighborCX,
                                                        unsigned int* neighborCY,
                                                        unsigned int* neighborCZ,
                                                        unsigned int* neighborFX,
                                                        unsigned int* neighborFY,
                                                        unsigned int* neighborFZ,
                                                        unsigned long long numberOfLBnodesC,
                                                        unsigned long long numberOfLBnodesF,
                                                        bool isEvenTimestep,
                                                        unsigned int* posC,
                                                        unsigned int* posFSWB,
                                                        unsigned int kFC,
                                                        real omCoarse,
                                                        real omFine,
                                                        real nu,
                                                        unsigned int nxC,
                                                        unsigned int nyC,
                                                        unsigned int nxF,
                                                        unsigned int nyF,
                                                        ICellNeigh neighborFineToCoarse);

__global__ void scaleFC_AA2016_comp_27(real* DC,
                                                  real* DF,
                                                  unsigned int* neighborCX,
                                                  unsigned int* neighborCY,
                                                  unsigned int* neighborCZ,
                                                  unsigned int* neighborFX,
                                                  unsigned int* neighborFY,
                                                  unsigned int* neighborFZ,
                                                  unsigned long long numberOfLBnodesC,
                                                  unsigned long long numberOfLBnodesF,
                                                  bool isEvenTimestep,
                                                  unsigned int* posC,
                                                  unsigned int* posFSWB,
                                                  unsigned int kFC,
                                                  real omCoarse,
                                                  real omFine,
                                                  real nu,
                                                  unsigned int nxC,
                                                  unsigned int nyC,
                                                  unsigned int nxF,
                                                  unsigned int nyF,
                                                  ICellNeigh neighborFineToCoarse);

__global__ void scaleFC_NSPress_27(real* DC,
                                              real* DF,
                                              unsigned int* neighborCX,
                                              unsigned int* neighborCY,
                                              unsigned int* neighborCZ,
                                              unsigned int* neighborFX,
                                              unsigned int* neighborFY,
                                              unsigned int* neighborFZ,
                                              unsigned long long numberOfLBnodesC,
                                              unsigned long long numberOfLBnodesF,
                                              bool isEvenTimestep,
                                              unsigned int* posC,
                                              unsigned int* posFSWB,
                                              unsigned int kFC,
                                              real omCoarse,
                                              real omFine,
                                              real nu,
                                              unsigned int nxC,
                                              unsigned int nyC,
                                              unsigned int nxF,
                                              unsigned int nyF,
                                              ICellNeigh neighborFineToCoarse);

__global__ void scaleFCThSMG7( real* DC,
                                          real* DF,
                                          real* DD7C,
                                          real* DD7F,
                                          unsigned int* neighborCX,
                                          unsigned int* neighborCY,
                                          unsigned int* neighborCZ,
                                          unsigned int* neighborFX,
                                          unsigned int* neighborFY,
                                          unsigned int* neighborFZ,
                                          unsigned long long numberOfLBnodesC,
                                          unsigned long long numberOfLBnodesF,
                                          bool isEvenTimestep,
                                          unsigned int* posC,
                                          unsigned int* posFSWB,
                                          unsigned int kFC,
                                          real nu,
                                          real diffusivity_coarse,
                                          ICellNeigh neighborFineToCoarse);

__global__ void scaleFCThS7(real* DC,
                                       real* DF,
                                       real* DD7C,
                                       real* DD7F,
                                       unsigned int* neighborCX,
                                       unsigned int* neighborCY,
                                       unsigned int* neighborCZ,
                                       unsigned int* neighborFX,
                                       unsigned int* neighborFY,
                                       unsigned int* neighborFZ,
                                       unsigned long long numberOfLBnodesC,
                                       unsigned long long numberOfLBnodesF,
                                       bool isEvenTimestep,
                                       unsigned int* posC,
                                       unsigned int* posFSWB,
                                       unsigned int kFC,
                                       real nu,
                                       real diffusivity_coarse);

__global__ void scaleFCThS27(  real* DC,
                                          real* DF,
                                          real* DD27C,
                                          real* DD27F,
                                          unsigned int* neighborCX,
                                          unsigned int* neighborCY,
                                          unsigned int* neighborCZ,
                                          unsigned int* neighborFX,
                                          unsigned int* neighborFY,
                                          unsigned int* neighborFZ,
                                          unsigned long long numberOfLBnodesC,
                                          unsigned long long numberOfLBnodesF,
                                          bool isEvenTimestep,
                                          unsigned int* posC,
                                          unsigned int* posFSWB,
                                          unsigned int kFC,
                                          real nu,
                                          real diffusivity_coarse,
                                          ICellNeigh neighborFineToCoarse);

__global__ void DragLiftPost27(  real* DD,
                                            int* k_Q,
                                            real* QQ,
                                            int numberOfBCnodes,
                                            double *DragX,
                                            double *DragY,
                                            double *DragZ,
                                            unsigned int* neighborX,
                                            unsigned int* neighborY,
                                            unsigned int* neighborZ,
                                            unsigned long long numberOfLBnodes,
                                            bool isEvenTimestep);

__global__ void DragLiftPre27(   real* DD,
                                            int* k_Q,
                                            real* QQ,
                                            int numberOfBCnodes,
                                            double *DragX,
                                            double *DragY,
                                            double *DragZ,
                                            unsigned int* neighborX,
                                            unsigned int* neighborY,
                                            unsigned int* neighborZ,
                                            unsigned long long numberOfLBnodes,
                                            bool isEvenTimestep);

__global__ void CalcCP27(real* DD,
                                    int* cpIndex,
                                    int nonCp,
                                    double *cpPress,
                                    unsigned int* neighborX,
                                    unsigned int* neighborY,
                                    unsigned int* neighborZ,
                                    unsigned long long numberOfLBnodes,
                                    bool isEvenTimestep);

__global__ void getSendFsPre27(real* DD,
                                          real* bufferFs,
                                          int* sendIndex,
                                          int buffmax,
                                          unsigned int* neighborX,
                                          unsigned int* neighborY,
                                          unsigned int* neighborZ,
                                          unsigned long long numberOfLBnodes,
                                          bool isEvenTimestep);

__global__ void getSendFsPost27(real* DD,
                                           real* bufferFs,
                                           int* sendIndex,
                                           int buffmax,
                                           unsigned int* neighborX,
                                           unsigned int* neighborY,
                                           unsigned int* neighborZ,
                                           unsigned long long numberOfLBnodes,
                                           bool isEvenTimestep);

__global__ void setRecvFsPre27(real* DD,
                                          real* bufferFs,
                                          int* recvIndex,
                                          int buffmax,
                                          unsigned int* neighborX,
                                          unsigned int* neighborY,
                                          unsigned int* neighborZ,
                                          unsigned long long numberOfLBnodes,
                                          bool isEvenTimestep);

__global__ void setRecvFsPost27(real* DD,
                                           real* bufferFs,
                                           int* recvIndex,
                                           int buffmax,
                                           unsigned int* neighborX,
                                           unsigned int* neighborY,
                                           unsigned int* neighborZ,
                                           unsigned long long numberOfLBnodes,
                                           bool isEvenTimestep);

__global__ void getSendGsF3(
    real* G6,
    real* bufferGs,
    int* sendIndex,
    int buffmax,
    unsigned int* neighborX,
    unsigned int* neighborY,
    unsigned int* neighborZ,
    unsigned long long numberOfLBnodes,
    bool isEvenTimestep);

__global__ void setRecvGsF3(
    real* G6,
    real* bufferGs,
    int* recvIndex,
    int buffmax,
    unsigned int* neighborX,
    unsigned int* neighborY,
    unsigned int* neighborZ,
    unsigned long long numberOfLBnodes,
    bool isEvenTimestep);

__global__ void WallFunction27(     real* vx,
                                            real* vy,
                                            real* vz,
                                            real* DD,
                                            int* k_Q,
                                            real* QQ,
                                            unsigned int numberOfBCnodes,
                                            real om1,
                                            unsigned int* neighborX,
                                            unsigned int* neighborY,
                                            unsigned int* neighborZ,
                                            unsigned long long numberOfLBnodes,
                                            bool isEvenTimestep);

__global__ void LBSetOutputWallVelocitySP27( real* vxD,
                                                        real* vyD,
                                                        real* vzD,
                                                        real* vxWall,
                                                        real* vyWall,
                                                        real* vzWall,
                                                        int numberOfWallNodes,
                                                        int* kWallNodes,
                                                        real* rhoD,
                                                        real* pressD,
                                                        unsigned int* geoD,
                                                        unsigned int* neighborX,
                                                        unsigned int* neighborY,
                                                        unsigned int* neighborZ,
                                                        unsigned long long numberOfLBnodes,
                                                        real* DD,
                                                        bool isEvenTimestep);

__global__ void CalcTurbulenceIntensity(
   real* vxx,
   real* vyy,
   real* vzz,
   real* vxy,
   real* vxz,
   real* vyz,
   real* vx_mean,
   real* vy_mean,
   real* vz_mean,
   real* DD,
   uint *typeOfGridNode,
   unsigned int* neighborX,
   unsigned int* neighborY,
   unsigned int* neighborZ,
   unsigned long long numberOfLBnodes,
   bool isEvenTimestep);

#endif
